@@ -1,8 +1,8 @@
 /*
- * $Id: action.c,v 1.142 2003/02/09 16:00:46 tal197 Exp $
+ * $Id: action.c,v 1.160 2005/01/22 15:20:10 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -51,6 +51,8 @@
 #include "dir.h"
 #include "icon.h"
 #include "mount.h"
+#include "type.h"
+#include "xtypes.h"
 
 /* Parent->Child messages are one character each:
  *
@@ -100,6 +102,7 @@ static unsigned long file_counter;	/* For Disk Usage */
 
 static struct mode_change *mode_change = NULL;	/* For Permissions */
 static FindCondition *find_condition = NULL;	/* For Find */
+static MIME_type *type_change = NULL;
 
 /* Only used by child */
 static gboolean o_force = FALSE;
@@ -117,6 +120,7 @@ static Option o_action_newer;
  */
 static guchar	*last_chmod_string = NULL;
 static guchar	*last_find_string = NULL;
+static guchar	*last_settype_string = NULL;
 
 /* Set to one of the above before forking. This may change over a call to
  * reply(). It is reset to NULL once the text is parsed.
@@ -128,7 +132,7 @@ static void send_done(void);
 static void send_check_path(const gchar *path);
 static void send_mount_path(const gchar *path);
 static gboolean printf_send(const char *msg, ...);
-static gboolean send(void);
+static gboolean send_msg(void);
 static gboolean send_error(void);
 static gboolean send_dir(const char *dir);
 static gboolean read_exact(int source, char *buffer, ssize_t len);
@@ -230,7 +234,7 @@ static void show_chmod_help(gpointer data)
 	GtkWidget *text;
 
 	help = gtk_dialog_new_with_buttons(
-			_("Find expression reference"),
+			_("Change permissions reference"),
 			NULL, 0,
 			GTK_STOCK_CLOSE, GTK_RESPONSE_CANCEL,
 			NULL);
@@ -275,6 +279,45 @@ static void show_chmod_help(gpointer data)
 	gtk_widget_show_all(help);
 }
 
+
+static void show_settype_help(gpointer data)
+{
+	GtkWidget *help;
+	GtkWidget *text;
+
+	help = gtk_dialog_new_with_buttons(
+			_("Set type reference"),
+			NULL, 0,
+			GTK_STOCK_CLOSE, GTK_RESPONSE_CANCEL,
+			NULL);
+	gtk_dialog_set_default_response(GTK_DIALOG(help), GTK_RESPONSE_CANCEL);
+
+	text = gtk_label_new(NULL);
+	gtk_misc_set_padding(GTK_MISC(text), 2, 2);
+	gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(help)->vbox), text);
+	gtk_label_set_selectable(GTK_LABEL(text), TRUE);
+	gtk_label_set_markup(GTK_LABEL(text), _(
+"Normally ROX-Filer determines the type of a regular file\n"
+"by matching it's name against a pattern. To change the\n"
+"type of the file you must rename it.\n" 
+"\n"
+"Newer file systems can support something called 'Extended\n"
+"Attributes' which can be used to store additional data with\n"
+"each file as named parameters. ROX-Filer uses the\n"
+"'user.mime_type' attribute to store file types.\n" 
+"\n"
+"File types are only supported for regular files, not\n"
+"directories, devices, pipes or sockets, and then only\n"
+"on certain file systems and where the OS implements them.\n" 
+"\n"
+ATTR_MAN_PAGE));
+
+	g_signal_connect(help, "response",
+			G_CALLBACK(gtk_widget_destroy), NULL);
+
+	gtk_widget_show_all(help);
+}
+
 static void process_message(GUIside *gui_side, const gchar *buffer)
 {
 	ABox *abox = gui_side->abox;
@@ -313,6 +356,17 @@ static void process_message(GUIside *gui_side, const gchar *buffer)
 		gui_side->errors++;
 		abox_log(abox, buffer + 1, "error");
 	}
+	else if (*buffer == '<') 
+		abox_set_file(abox, 0, buffer+1);
+	else if (*buffer == '>')
+	{
+		abox_set_file(abox, 1, buffer+1);
+		abox_show_compare(abox, TRUE);
+	}
+	else if (*buffer == '%')
+	{
+		abox_set_percentage(abox, atoi(buffer+1));
+	}
 	else
 		abox_log(abox, buffer + 1, NULL);
 }
@@ -463,11 +517,11 @@ static gboolean printf_send(const char *msg, ...)
 	g_string_assign(message, tmp);
 	g_free(tmp);
 
-	return send();
+	return send_msg();
 }
 
 /* Send 'message' to our parent process. TRUE on success. */
-static gboolean send(void)
+static gboolean send_msg(void)
 {
 	char len_buffer[5];
 	ssize_t len;
@@ -508,6 +562,7 @@ static void response(GtkDialog *dialog, gint response, GUIside *gui_side)
 
 	fputc(code, gui_side->to_child);
 	fflush(gui_side->to_child);
+	abox_show_compare(gui_side->abox, FALSE);
 }
 
 static void flag_toggled(ABox *abox, gint flag, GUIside *gui_side)
@@ -607,7 +662,7 @@ static void check_flags(void)
 
 /* Read until the user sends a reply. If ignore_quiet is TRUE then
  * the user MUST click Yes or No, else treat quiet on as Yes.
- * If the user needs prompting then does send().
+ * If the user needs prompting then does send_msg().
  */
 static gboolean printf_reply(int fd, gboolean ignore_quiet,
 			     const char *msg, ...)
@@ -627,7 +682,7 @@ static gboolean printf_reply(int fd, gboolean ignore_quiet,
 	g_string_assign(message, tmp);
 	g_free(tmp);
 
-	send();
+	send_msg();
 
 	while (1)
 	{
@@ -790,10 +845,10 @@ static GUIside *start_action(GtkWidget *abox, ActionChild *func, gpointer data,
 	g_signal_connect(abox, "abort_operation",
 			 G_CALLBACK(abort_operation), gui_side);
 
-	gui_side->input_tag = gtk_input_add_full(gui_side->from_child,
+	gui_side->input_tag = gdk_input_add_full(gui_side->from_child,
 						GDK_INPUT_READ,
 						message_from_child,
-						NULL, gui_side, NULL);
+						gui_side, NULL);
 
 	return gui_side;
 }
@@ -854,10 +909,16 @@ static void do_delete(const char *src_path, const char *unused)
 					   : access(src_path, W_OK) != 0;
 	if (write_prot || !quiet)
 	{
-		if (!printf_reply(from_parent, write_prot && !o_force,
+		int res;
+		
+		printf_send("<%s", src_path);
+		printf_send(">");
+		res=printf_reply(from_parent, write_prot && !o_force,
 				  _("?Delete %s'%s'?"),
 				  write_prot ? _("WRITE-PROTECTED ") : "",
-				  src_path))
+				 src_path);
+		printf_send("<");
+		if (!res)
 			return;
 	}
 	else if (!o_brief)
@@ -894,6 +955,53 @@ static void do_delete(const char *src_path, const char *unused)
 	g_free(safe_path);
 }
 
+static void do_eject(const char *path)
+{
+	const char *argv[3] = {NULL, NULL, NULL};
+	char *err;
+	
+	check_flags();
+
+	if (!quiet)
+	{
+		int res;
+		printf_send("<%s", path);
+		printf_send(">");
+		res=printf_reply(from_parent, !o_force,
+				  _("?Eject '%s'?"),
+				 path);
+		printf_send("<");
+		if (!res)
+			return;
+	}
+	else if (!o_brief)
+		printf_send(_("'Eject '%s'\n"), path);
+
+	/* Need to close all sub-directories now, or we
+	 * can't unmount if dnotify is used.
+	 */
+	{
+		char c = '?';
+		printf_send("X%s", path);
+		/* Wait until it's safe... */
+		read(from_parent, &c, 1);
+		g_return_if_fail(c == 'X');
+	}
+
+	argv[0] = "eject";
+	argv[1] = path;
+	argv[2] = NULL;
+	err = fork_exec_wait(argv);
+	if (err)
+	{
+		printf_send(_("!%s\neject failed\n"), err);
+		g_free(err);
+	}
+
+	printf_send("M%s", path);
+
+}
+
 /* path is the item to check. If is is a directory then we may recurse
  * (unless prune is used).
  */
@@ -1001,8 +1109,13 @@ static void do_chmod(const char *path, const char *unused)
 
 	if (!quiet)
 	{
-		if (!printf_reply(from_parent, FALSE,
-				  _("?Change permissions of '%s'?"), path))
+		int res;
+		printf_send("<%s", path);
+		printf_send(">");
+		res=printf_reply(from_parent, FALSE,
+				 _("?Change permissions of '%s'?"), path);
+		printf_send("<");
+		if (!res)
 			return;
 	}
 	else if (!o_brief)
@@ -1060,6 +1173,89 @@ static void do_chmod(const char *path, const char *unused)
 	}
 }
 
+static void do_settype(const char *path, const char *unused)
+{
+	struct stat 	info;
+
+	check_flags();
+
+	if (mc_lstat(path, &info))
+	{
+		send_error();
+		return;
+	}
+	if (S_ISLNK(info.st_mode))
+		return;
+
+	if (!quiet)
+	{
+		int res;
+		printf_send("<%s", path);
+		printf_send(">");
+		res=printf_reply(from_parent, FALSE,
+				 _("?Change type of '%s'?"), path);
+		printf_send("<");
+		if (!res)
+			return;
+	}
+
+	for (;;)
+	{
+		if (new_entry_string)
+		{
+			type_change = mime_type_lookup(new_entry_string);
+			null_g_free(&new_entry_string);
+		}
+
+		if (type_change)
+			break;
+
+		printf_send(_("!Invalid type - "
+			      "change it and try again\n"));
+		if (!printf_reply(from_parent, TRUE,
+				  _("?Change type of '%s'?"), path))
+			return;
+	}
+
+	if (mc_lstat(path, &info))
+	{
+		send_error();
+		return;
+	}
+	if (S_ISLNK(info.st_mode))
+		return;
+
+	if (S_ISREG(info.st_mode))
+	{
+		if (!o_brief)
+		{
+			const char *comment;
+
+			comment = mime_type_comment(type_change);
+			printf_send(_("'Changing type of '%s' to '%s'\n"), path,
+				    comment);
+		}
+
+		if (xtype_set(path, type_change))
+		{
+			send_error();
+			return;
+		}
+
+		send_check_path(path);
+	}
+	else if (S_ISDIR(info.st_mode))
+	{
+		if (o_recurse)
+		{
+			guchar *safe_path;
+			safe_path = g_strdup(path);
+			for_dir_contents(do_settype, safe_path, unused);
+			g_free(safe_path);
+		}
+	}
+}
+
 /* We want to copy 'object' into directory 'dir'. If 'action_leaf'
  * is set then that is the new leafname, otherwise the leafname stays
  * the same.
@@ -1112,6 +1308,8 @@ static void do_copy2(const char *path, const char *dest)
 		}
 		else
 		{
+			printf_send("<%s", path);
+			printf_send(">%s", dest_path);
 			if (!printf_reply(from_parent, TRUE,
 					  _("?'%s' already exists - %s?"),
 					  dest_path,
@@ -1138,11 +1336,13 @@ static void do_copy2(const char *path, const char *dest)
 	}
 	else if (!quiet)
 	{
+		printf_send("<%s", path);
+		printf_send(">");
 		if (!printf_reply(from_parent, FALSE,
 				  _("?Copy %s as %s?"), path, dest_path))
 			return;
 	}
-	else
+	else if (!o_brief || S_ISDIR(info.st_mode))
 		printf_send(_("'Copying %s as %s\n"), path, dest_path);
 
 	if (S_ISDIR(info.st_mode))
@@ -1269,10 +1469,15 @@ static void do_move2(const char *path, const char *dest)
 		{
 			/* Newer; keep going */
 		}
-		else if (!printf_reply(from_parent, TRUE,
+		else
+		{
+			printf_send("<%s", path);
+			printf_send(">%s", dest_path);
+			if (!printf_reply(from_parent, TRUE,
 				       _("?'%s' already exists - overwrite?"),
 				       dest_path))
-			return;
+				return;
+		}
 
 		if (S_ISDIR(info.st_mode))
 			err = rmdir(dest_path);
@@ -1289,11 +1494,13 @@ static void do_move2(const char *path, const char *dest)
 	}
 	else if (!quiet)
 	{
+		printf_send("<%s", path);
+		printf_send(">");
 		if (!printf_reply(from_parent, FALSE,
 				  _("?Move %s as %s?"), path, dest_path))
 			return;
 	}
-	else
+	else if (!o_brief)
 		printf_send(_("'Moving %s as %s\n"), path, dest_path);
 
 	argv[2] = path;
@@ -1356,9 +1563,13 @@ static void do_link(const char *path, const char *dest)
 
 	if (quiet)
 		printf_send(_("'Linking %s as %s\n"), path, dest_path);
-	else if (!printf_reply(from_parent, FALSE,
-			       _("?Link %s as %s?"), path, dest_path))
-		return;
+	else {
+		printf_send("<%s", path);
+		printf_send(">");
+		if (!printf_reply(from_parent, FALSE,
+				  _("?Link %s as %s?"), path, dest_path))
+			return;
+	}
 
 	if (symlink(path, dest_path))
 		send_error();
@@ -1406,13 +1617,20 @@ static void do_mount(const guchar *path, gboolean mount)
 			_("!%s\nMount failed\n") :
 			_("!%s\nUnmount failed\n"), err);
 		g_free(err);
+
+		/* Mount may have worked even on error, eg if we try to mount
+		 * a read-only disk read/write, it gets mounted read-only
+		 * with an error.
+		 */
+		if (mount && mount_is_mounted(path, NULL, NULL))
+			printf_send(_("'(seems to be mounted now anyway)\n"));
+		else
+			return;
 	}
-	else
-	{
-		printf_send("M%s", path);
-		if (mount && mount_open_dir)
-			printf_send("o%s", path);
-	}
+
+	printf_send("M%s", path);
+	if (mount && mount_open_dir)
+		printf_send("o%s", path);
 }
 
 /*			CHILD MAIN LOOPS			*/
@@ -1424,10 +1642,12 @@ static void usage_cb(gpointer data)
 {
 	GList *paths = (GList *) data;
 	double	total_size = 0;
+	int n, i, per;
 
+	n=g_list_length(paths);
 	dir_counter = file_counter = 0;
 
-	for (; paths; paths = paths->next)
+	for (i=0; paths; paths = paths->next, i++)
 	{
 		guchar	*path = (guchar *) paths->data;
 
@@ -1435,6 +1655,11 @@ static void usage_cb(gpointer data)
 
 		size_tally = 0;
 		
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
 		do_usage(path, NULL);
 
 		printf_send("'%s: %s\n",
@@ -1460,7 +1685,7 @@ static void usage_cb(gpointer data)
 				dir_counter == 1 ? _("directory")
 						 : _("directories"));
 	
-	send();
+	send_msg();
 }
 
 #ifdef DO_MOUNT_POINTS
@@ -1468,16 +1693,23 @@ static void mount_cb(gpointer data)
 {
 	GList 		*paths = (GList *) data;
 	gboolean	mount_points = FALSE;
+	int n, i, per;
 
-	for (; paths; paths = paths->next)
+	n=g_list_length(paths);
+	for (i=0; paths; paths = paths->next, i++)
 	{
 		guchar *path = (guchar *) paths->data;
 		guchar *target;
 
-		target = readlink_dup(path);
+		target = pathdup(path);
 		if (!target)
 			target = path;
 
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
 		if (mount_is_mounted(target, NULL, NULL))
 		{
 			mount_points = TRUE;
@@ -1516,15 +1748,22 @@ static guchar *dirname(guchar *path)
 static void delete_cb(gpointer data)
 {
 	GList	*paths = (GList *) data;
+	int n, i, per;
 
-	for (; paths; paths = paths->next)
+	n=g_list_length(paths);
+	for (i=0; paths; paths = paths->next, i++)
 	{
 		guchar	*path = (guchar *) paths->data;
 		guchar	*dir;
-		
+
 		dir = dirname(path);
 		send_dir(dir);
 
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
 		do_delete(path, dir);
 
 		g_free(dir);
@@ -1533,6 +1772,30 @@ static void delete_cb(gpointer data)
 	send_done();
 }
 
+static void eject_cb(gpointer data)
+{
+	GList	*paths = (GList *) data;
+	int n, i, per;
+
+	n=g_list_length(paths);
+
+	for (i=0; paths; paths = paths->next, i++)
+	{
+		guchar	*path = (guchar *) paths->data;
+		
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
+		send_dir(path);
+
+		do_eject(path);
+	}
+	
+	send_done();
+}
+
 static void find_cb(gpointer data)
 {
 	GList *all_paths = (GList *) data;
@@ -1561,12 +1824,20 @@ static void find_cb(gpointer data)
 static void chmod_cb(gpointer data)
 {
 	GList *paths = (GList *) data;
+	int n, i, per;
 
-	for (; paths; paths = paths->next)
+	n=g_list_length(paths);
+
+	for (i=0; paths; paths = paths->next, i++)
 	{
 		guchar	*path = (guchar *) paths->data;
 		struct stat info;
 
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
 		send_dir(path);
 
 		if (mc_stat(path, &info) != 0)
@@ -1581,12 +1852,51 @@ static void chmod_cb(gpointer data)
 	send_done();
 }
 
+static void settype_cb(gpointer data)
+{
+	GList *paths = (GList *) data;
+	int n, i, per;
+
+	n=g_list_length(paths);
+
+	for (i=0; paths; paths = paths->next, i++)
+	{
+		guchar	*path = (guchar *) paths->data;
+		struct stat info;
+
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
+		send_dir(path);
+
+		if (mc_stat(path, &info) != 0)
+			send_error();
+		else if (S_ISLNK(info.st_mode))
+			printf_send(_("!'%s' is a symbolic link\n"),
+				    g_basename(path));
+		else
+			do_settype(path, NULL);
+	}
+	
+	send_done();
+}
+
 static void list_cb(gpointer data)
 {
 	GList	*paths = (GList *) data;
+	int n, i, per;
 
-	for (; paths; paths = paths->next)
+	n=g_list_length(paths);
+
+	for (i=0; paths; paths = paths->next, i++)
 	{
+		if(n>1 && i>0)
+		{
+			per=100*i/n;
+			printf_send("%%%d", per);
+		}
 		send_dir((char *) paths->data);
 
 		action_do_func((char *) paths->data, action_dest);
@@ -1636,7 +1946,7 @@ void action_find(GList *paths)
 	gui_side->entry_string_func = set_find_string_colour;
 
 	number_of_windows++;
-	gtk_widget_show_all(abox);
+	gtk_widget_show(abox);
 }
 
 /* Count disk space used by selected items */
@@ -1786,7 +2096,7 @@ void action_chmod(GList *paths, gboolean force_recurse, const char *action)
 		'R', recurse);
 
 	gui_side->default_string = &last_chmod_string;
-	abox_add_combo(ABOX(abox), presets, new_entry_string,
+	abox_add_combo(ABOX(abox), _("Command:"), presets, new_entry_string,
 				new_help_button(show_chmod_help, NULL));
 
 	g_signal_connect(ABOX(abox)->entry, "changed",
@@ -1804,6 +2114,64 @@ out:
 	null_g_free(&new_entry_string);
 }
 
+/* Set the MIME type of the selected items */
+void action_settype(GList *paths, gboolean force_recurse, const char *oldtype)
+{
+	GtkWidget	*abox;
+	GUIside		*gui_side;
+	GList		*presets = NULL;
+	gboolean	recurse = force_recurse || o_action_recurse.int_value;
+
+	if (!paths)
+	{
+		report_error(_("You need to select the items "
+				"whose type you want to change"));
+		return;
+	}
+
+	if (!last_settype_string)
+		last_settype_string = g_strdup("text/plain");
+
+	if (oldtype)
+		new_entry_string = g_strdup(oldtype);
+	else
+		new_entry_string = g_strdup(last_settype_string);
+
+	abox = abox_new(_("Set type"), FALSE);
+	gui_side = start_action(abox, settype_cb, paths,
+				o_action_force.int_value,
+				o_action_brief.int_value,
+				recurse,
+				o_action_newer.int_value);
+	
+	if (!gui_side)
+		goto out;
+
+	abox_add_flag(ABOX(abox),
+		_("Brief"), _("Don't list processed files"),
+		'B', o_action_brief.int_value);
+	abox_add_flag(ABOX(abox),
+		_("Recurse"), _("Change contents of subdirectories"),
+		'R', recurse);
+
+	gui_side->default_string = &last_settype_string;
+
+	/* Note: get the list again each time -- it can change */
+	presets = mime_type_name_list();
+	abox_add_combo(ABOX(abox), _("Type:"), presets, new_entry_string,
+				new_help_button(show_settype_help, NULL));
+	g_list_free(presets);
+
+	g_signal_connect(ABOX(abox)->entry, "changed",
+			G_CALLBACK(entry_changed), gui_side);
+
+	number_of_windows++;
+	gtk_widget_show(abox);
+
+out:
+	null_g_free(&new_entry_string);
+}
+
 /* If leaf is NULL then the copy has the same name as the original.
  * quiet can be -1 for default.
  */
@@ -1832,6 +2200,9 @@ void action_copy(GList *paths, const char *dest, const char *leaf, int quiet)
 		   _("Newer"),
 		   _("Only over-write if source is newer than destination."),
 		   'W', o_action_newer.int_value);
+	abox_add_flag(ABOX(abox),
+		_("Brief"), _("Only log directories as they are copied"),
+		'B', o_action_brief.int_value);
 
 	number_of_windows++;
 	gtk_widget_show(abox);
@@ -1865,6 +2236,9 @@ void action_move(GList *paths, const char *dest, const char *leaf, int quiet)
 		   _("Newer"),
 		   _("Only over-write if source is newer than destination."),
 		   'W', o_action_newer.int_value);
+	abox_add_flag(ABOX(abox),
+		_("Brief"), _("Don't log each file as it is moved"),
+		'B', o_action_brief.int_value);
 	number_of_windows++;
 	gtk_widget_show(abox);
 }
@@ -1893,6 +2267,25 @@ void action_link(GList *paths, const char *dest, const char *leaf)
 	gtk_widget_show(abox);
 }
 
+/* Eject these paths */
+void action_eject(GList *paths)
+{
+	GUIside		*gui_side;
+	GtkWidget	*abox;
+
+	abox = abox_new(_("Eject"), TRUE);
+	gui_side = start_action(abox, eject_cb, paths,
+					 o_action_force.int_value,
+					 o_action_brief.int_value,
+					 o_action_recurse.int_value,
+					 o_action_newer.int_value);
+	if (!gui_side)
+		return;
+
+	number_of_windows++;
+	gtk_widget_show(abox);
+}
+
 void action_init(void)
 {
 	option_add_int(&o_action_copy, "action_copy", 1);
