@@ -1,8 +1,8 @@
 /*
- * $Id: filer.c,v 1.339 2003/10/26 16:14:28 tal197 Exp $
+ * $Id: filer.c,v 1.357 2005/01/16 16:43:30 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -30,6 +30,7 @@
 #include <ctype.h>
 #include <netdb.h>
 #include <sys/param.h>
+#include <fnmatch.h>
 
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
@@ -37,12 +38,12 @@
 
 #include "global.h"
 
+#include "filer.h"
 #include "display.h"
 #include "main.h"
 #include "fscache.h"
 #include "support.h"
 #include "gui_support.h"
-#include "filer.h"
 #include "choices.h"
 #include "pixmaps.h"
 #include "menu.h"
@@ -86,6 +87,39 @@ static GHashTable *window_with_id = NULL;
 
 static FilerWindow *window_with_primary = NULL;
 
+static GHashTable *settings_table=NULL;
+	
+typedef struct {
+	gchar *path;
+
+	guint flags;   /* Which parts are valid, see below */
+
+	gint x, y;
+	gint width, height;
+	gboolean show_hidden;
+	ViewType view_type;
+	SortType	sort_type;
+	GtkSortType	sort_order;
+	gboolean        show_thumbs;
+
+	DetailsType	details_type;
+	DisplayStyle	display_style;
+
+	FilterType filter_type;
+	char *filter;
+} Settings;
+
+enum settings_flags{
+	SET_POSITION=1,   /* x, y */
+	SET_SIZE=2,       /* width, height */
+	SET_HIDDEN=4,     /* show_hidden */
+	SET_STYLE=8,      /* display_style */
+	SET_SORT=16,      /* sort_type, sort_order */
+	SET_DETAILS=32,   /* view_type, details_type */
+	SET_THUMBS=64,    /* show_thumbs */
+	SET_FILTER=128,   /* filter_type, filter */
+};
+
 /* Static prototypes */
 static void attach(FilerWindow *filer_window);
 static void detach(FilerWindow *filer_window);
@@ -119,6 +153,10 @@ static gboolean drag_motion(GtkWidget		*widget,
                             guint		time,
 			    FilerWindow		*filer_window);
 
+static void load_settings(void);
+static void save_settings(void);
+static void check_settings(FilerWindow *filer_window);
+
 static GdkCursor *busy_cursor = NULL;
 static GdkCursor *crosshair = NULL;
 
@@ -179,6 +217,8 @@ void filer_init(void)
 	}
 	
 	g_free(dpyhost);
+
+	load_settings();
 }
 
 static gboolean if_deleted(gpointer item, gpointer removed)
@@ -196,6 +236,8 @@ static gboolean if_deleted(gpointer item, gpointer removed)
 	return FALSE;
 }
 
+#define DECOR_BORDER 32
+
 /* Resize the filer window to w x h pixels, plus border (not clamped).
  * If triggered by a key event, warp the pointer (for SloppyFocus users).
  */
@@ -218,20 +260,34 @@ void filer_window_set_size(FilerWindow *filer_window, int w, int h)
 
 	if (GTK_WIDGET_VISIBLE(window))
 	{
-		gint x, y;
+		gint x, y, m;
 		GtkRequisition	*req = &window->requisition;
 		GdkWindow *gdk_window = window->window;
 
 		w = MAX(req->width, w);
 		h = MAX(req->height, h);
+		gdk_window_get_pointer(NULL, &x, &y, NULL);
+		m = gdk_screen_get_monitor_at_point(gdk_screen_get_default(),
+				x, y);
 		gdk_window_get_position(gdk_window, &x, &y);
 
-		if (x + w > screen_width || y + h > screen_height)
+		if (x + w + DECOR_BORDER >
+				monitor_geom[m].x + monitor_geom[m].width ||
+				y + h + DECOR_BORDER >
+				monitor_geom[m].y + monitor_geom[m].height)
 		{
-			if (x + w > screen_width)
-				x = screen_width - w - 4;
-			if (y + h > screen_height)
-				y = screen_height - h - 4;
+			if (x + w + DECOR_BORDER >
+				monitor_geom[m].x + monitor_geom[m].width)
+			{
+				x = monitor_geom[m].x + monitor_geom[m].width -
+					w - 4 - DECOR_BORDER;
+			}
+			if (y + h + DECOR_BORDER >
+				monitor_geom[m].y + monitor_geom[m].height)
+			{
+				y = monitor_geom[m].y + monitor_geom[m].height -
+					h - 4 - DECOR_BORDER;
+			}
 			gdk_window_move_resize(gdk_window, x, y, w, h);
 		}
 		else
@@ -243,13 +299,25 @@ void filer_window_set_size(FilerWindow *filer_window, int w, int h)
 	event = gtk_get_current_event();
 	if (event && event->type == GDK_KEY_PRESS)
 	{
+		int x, y;
+		int nx, ny;
+
 		GdkWindow *win = filer_window->window->window;
+
+		gdk_window_get_pointer(filer_window->window->window,
+					&x, &y, NULL);
+
+		nx = CLAMP(x, 4, w - 4);
+		ny = CLAMP(y, 4, h - 4);
 		
-		XWarpPointer(gdk_x11_drawable_get_xdisplay(win),
-			     None,
-			     gdk_x11_drawable_get_xid(win),
-			     0, 0, 0, 0,
-			     w - 4, h - 4);
+		if (nx != x || ny != y)
+		{
+			XWarpPointer(gdk_x11_drawable_get_xdisplay(win),
+					None,
+					gdk_x11_drawable_get_xid(win),
+					0, 0, 0, 0,
+					nx, ny);
+		}
 	}
 }
 
@@ -258,23 +326,43 @@ void filer_window_set_size(FilerWindow *filer_window, int w, int h)
  */
 static gint open_filer_window(FilerWindow *filer_window)
 {
+	Settings	*dir_settings = (Settings *) g_hash_table_lookup(settings_table,
+					      filer_window->sym_path);
+	gboolean force_resize = !(o_filer_auto_resize.int_value == RESIZE_NEVER && dir_settings && dir_settings->flags & SET_POSITION);
+
 	view_style_changed(filer_window->view, 0);
 
 	if (filer_window->open_timeout)
 	{
-		gtk_timeout_remove(filer_window->open_timeout);
+		g_source_remove(filer_window->open_timeout);
 		filer_window->open_timeout = 0;
 	}
 
 	if (!GTK_WIDGET_VISIBLE(filer_window->window))
 	{
-		display_set_actual_size(filer_window, TRUE);
+		display_set_actual_size(filer_window, force_resize);
 		gtk_widget_show(filer_window->window);
 	}
 
 	return FALSE;
 }
 
+/* Look through all items we want to display, and queue a recheck on any
+ * that require it.
+ */
+static void queue_interesting(FilerWindow *filer_window)
+{
+	DirItem	*item;
+	ViewIter iter;
+
+	view_get_iter(filer_window->view, &iter, 0);
+	while ((item = iter.next(&iter)))
+	{
+		if (item->flags & ITEM_FLAG_NEED_RESCAN_QUEUE)
+			dir_queue_recheck(filer_window->directory, item);
+	}
+}
+
 static void update_display(Directory *dir,
 			DirAction	action,
 			GPtrArray	*items,
@@ -332,6 +420,9 @@ static void update_display(Directory *dir,
 		case DIR_ERROR_CHANGED:
 			filer_set_title(filer_window);
 			break;
+		case DIR_QUEUE_INTERESTING:
+			queue_interesting(filer_window);
+			break;
 	}
 }
 
@@ -368,6 +459,121 @@ static void detach(FilerWindow *filer_window)
 	filer_window->directory = NULL;
 }
 
+/* If 'start' was mounted by ROX-Filer, return it. Otherwise, try the
+ * parents up the tree.
+ * NULL if we're not in a user mount point.
+ * g_free() the result.
+ */
+static char *get_ancestor_user_mount_point(const char *start)
+{
+	char *path;
+
+	path = strdup(start);
+
+	while (1)
+	{
+		char *slash;
+
+		if (mount_is_user_mounted(path))
+			return path;
+
+		if (!path[1])
+		{
+			g_free(path);
+			return NULL;
+		}
+
+		slash = strrchr(path + 1, '/');
+		if (!slash)
+			slash = path + 1;
+		*slash = '\0';
+	}
+}
+
+static void umount_dialog_response(GtkWidget *dialog, int response, char *mount)
+{
+	if (response == GTK_RESPONSE_OK)
+	{
+		GList *list; 
+
+		list = g_list_prepend(NULL, mount);
+		action_mount(list, FALSE, TRUE);
+		g_list_free(list);
+	}
+
+	g_free(mount);
+
+	gtk_widget_destroy(dialog);
+
+	one_less_window();
+}
+
+/* 'filer_window' shows a directory under 'mount'. If no other window also
+ * shows a directory under it, display a non-modal dialog offering to
+ * unmount the directory.
+ * 'mount' is freed by this function, either directly, or after the dialog
+ * closes.
+ */
+static void may_offer_unmount(FilerWindow *filer_window, char *mount)
+{
+	GtkWidget *dialog, *button;
+	GList	*next;
+	int len;
+	
+	len = strlen(mount);
+
+	for (next = all_filer_windows; next; next = next->next)
+	{
+		FilerWindow *other = (FilerWindow *) next->data;
+
+		if (other == filer_window)
+			continue;
+
+		if (strncmp(filer_window->real_path, other->real_path,
+			    len) != 0)
+			continue;
+
+		g_return_if_fail(
+			filer_window->real_path[len] != '/' ||
+			filer_window->real_path[len] != '\0');
+
+		if (other->real_path[len] != '/' &&
+		    other->real_path[len] != '\0')
+			continue;
+
+		/* Found another window. Don't offer to unmount. */
+		g_free(mount);
+		return;
+	}
+
+	dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION,
+			GTK_BUTTONS_NONE, 
+			_("Do you want to unmount this device?\n\n"
+			"Unmounting a device makes it safe to remove "
+			"the disk."), mount);
+
+	button = button_new_mixed(ROX_STOCK_MOUNTED, _("No change"));
+	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
+	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), button,
+					GTK_RESPONSE_CANCEL);
+	gtk_widget_show(button);
+
+	button = button_new_mixed(ROX_STOCK_MOUNT, _("Unmount"));
+	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
+	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), button,
+					GTK_RESPONSE_OK);
+	gtk_widget_show(button);
+
+	g_signal_connect(G_OBJECT(dialog), "response",
+			G_CALLBACK(umount_dialog_response), mount);
+
+	gtk_dialog_set_default_response(GTK_DIALOG(dialog),
+			GTK_RESPONSE_OK);
+
+	number_of_windows++;
+	gtk_widget_show(dialog);
+}
+
 /* Returns TRUE to prevent closing the window. May offer to unmount a
  * device.
  */
@@ -375,32 +581,12 @@ gboolean filer_window_delete(GtkWidget *window,
 			     GdkEvent *unused, /* (may be NULL) */
 			     FilerWindow *filer_window)
 {
-	if (mount_is_user_mounted(filer_window->real_path))
-	{
-		int action;
+	char *mount;
 
-		action = get_choice(PROJECT,
-			_("Do you want to unmount this device?\n\n"
-			"Unmounting a device makes it safe to remove "
-			"the disk."), 3,
-			GTK_STOCK_CANCEL, NULL,
-			GTK_STOCK_CLOSE, NULL,
-			ROX_STOCK_MOUNT, _("Unmount"));
+	mount = get_ancestor_user_mount_point(filer_window->real_path);
 
-		if (action == 0)
-			return TRUE;	/* Cancel close operation */
-
-		if (action == 2)
-		{
-			GList *list; 
-
-			list = g_list_prepend(NULL, filer_window->sym_path);
-			action_mount(list, FALSE, TRUE);
-			g_list_free(list);
-
-			return TRUE;
-		}
-	}
+	if (mount)
+		may_offer_unmount(filer_window, mount);
 
 	return FALSE;
 }
@@ -425,13 +611,13 @@ static void filer_window_destroyed(GtkWidget *widget, FilerWindow *filer_window)
 
 	if (filer_window->open_timeout)
 	{
-		gtk_timeout_remove(filer_window->open_timeout);
+		g_source_remove(filer_window->open_timeout);
 		filer_window->open_timeout = 0;
 	}
 
 	if (filer_window->auto_scroll != -1)
 	{
-		gtk_timeout_remove(filer_window->auto_scroll);
+		g_source_remove(filer_window->auto_scroll);
 		filer_window->auto_scroll = -1;
 	}
 
@@ -442,6 +628,11 @@ static void filer_window_destroyed(GtkWidget *widget, FilerWindow *filer_window)
 
 	filer_set_id(filer_window, NULL);
 
+	if(filer_window->filter_string)
+		g_free(filer_window->filter_string);
+	if(filer_window->regexp)
+		g_free(filer_window->regexp);
+
 	g_free(filer_window->auto_select);
 	g_free(filer_window->real_path);
 	g_free(filer_window->sym_path);
@@ -543,12 +734,12 @@ static void selection_get(GtkWidget *widget,
 	}
 	
 	if (reply->len > 0)
-		gtk_selection_data_set(selection_data, xa_string,
-				8, reply->str + 1, reply->len - 1);
+		gtk_selection_data_set_text(selection_data,
+				reply->str + 1, reply->len - 1);
 	else
 	{
 		g_warning("Attempt to paste empty selection!");
-		gtk_selection_data_set(selection_data, xa_string, 8, "", 0);
+		gtk_selection_data_set_text(selection_data, "", 0);
 	}
 
 	g_string_free(reply, TRUE);
@@ -605,7 +796,7 @@ void filer_openitem(FilerWindow *filer_window, ViewIter *iter, OpenFlags flags)
 		return;
 	}
 
-	if (!item->image)
+	if (item->base_type == TYPE_UNKNOWN)
 		dir_update_item(filer_window->directory, item->leafname);
 
 	if (item->base_type == TYPE_DIRECTORY)
@@ -661,7 +852,7 @@ static gint pointer_out(GtkWidget *widget,
 /* Move the cursor to the next selected item in direction 'dir'
  * (+1 or -1).
  */
-static void next_selected(FilerWindow *filer_window, int dir)
+void filer_next_selected(FilerWindow *filer_window, int dir)
 {
 	ViewIter	iter, cursor;
 	gboolean	have_cursor;
@@ -942,10 +1133,10 @@ gint filer_key_press_event(GtkWidget	*widget,
 			return_pressed(filer_window, event);
 			break;
 		case GDK_ISO_Left_Tab:
-			next_selected(filer_window, -1);
+			filer_next_selected(filer_window, -1);
 			break;
 		case GDK_Tab:
-			next_selected(filer_window, 1);
+			filer_next_selected(filer_window, 1);
 			break;
 		case GDK_BackSpace:
 			change_to_parent(filer_window);
@@ -1008,6 +1199,10 @@ void change_to_parent(FilerWindow *filer_window)
 
 	if (current[0] == '/' && current[1] == '\0')
 		return;		/* Already in the root */
+
+	if (mount_is_user_mounted(filer_window->real_path))
+		may_offer_unmount(filer_window,
+				g_strdup(filer_window->real_path));
 	
 	dir = g_path_get_dirname(current);
 	filer_change_to(filer_window, dir, g_basename(current));
@@ -1094,14 +1289,15 @@ void filer_change_to(FilerWindow *filer_window,
 
 	attach(filer_window);
 	
+	check_settings(filer_window);
+
 	display_set_actual_size(filer_window, FALSE);
 
 	if (o_filer_auto_resize.int_value == RESIZE_ALWAYS)
 		view_autosize(filer_window->view);
 
 	if (filer_window->mini_type == MINI_PATH)
-		gtk_idle_add((GtkFunction) minibuffer_show_cb,
-				filer_window);
+		g_idle_add((GSourceFunc) minibuffer_show_cb, filer_window);
 }
 
 /* Returns a list containing the full (sym) pathname of every selected item.
@@ -1153,6 +1349,8 @@ FilerWindow *filer_opendir(const char *path, FilerWindow *src_win,
 	DetailsType     dtype;
 	SortType	s_type;
 	GtkSortType	s_order;
+	Settings	*dir_settings = NULL;
+	gboolean 	force_resize = TRUE;
 
 	/* Get the real pathname of the directory and copy it */
 	real_path = pathdup(path);
@@ -1218,6 +1416,10 @@ FilerWindow *filer_opendir(const char *path, FilerWindow *src_win,
 	filer_window->max_thumbs = 0;
 	filer_window->sort_type = -1;
 
+	filer_window->filter = FILER_SHOW_ALL;
+	filer_window->filter_string = NULL;
+	filer_window->regexp = NULL;
+	
 	if (src_win && o_display_inherit_options.int_value)
 	{
 		s_type = src_win->sort_type;
@@ -1227,6 +1429,9 @@ FilerWindow *filer_opendir(const char *path, FilerWindow *src_win,
 		filer_window->show_hidden = src_win->show_hidden;
 		filer_window->show_thumbs = src_win->show_thumbs;
 		filer_window->view_type = src_win->view_type;
+
+		filer_set_filter(filer_window, src_win->filter,
+				 src_win->filter_string);
 	}
 	else
 	{
@@ -1239,24 +1444,74 @@ FilerWindow *filer_opendir(const char *path, FilerWindow *src_win,
 		filer_window->view_type = o_filer_view_type.int_value;
 	}
 
+	dir_settings = (Settings *) g_hash_table_lookup(settings_table,
+					      filer_window->sym_path);
+	if (dir_settings)
+	{
+		/* Override the current defaults with the per-directory
+		 * user settings.
+		 */
+		if (dir_settings->flags & SET_HIDDEN)
+			filer_window->show_hidden = dir_settings->show_hidden;
+
+		if (dir_settings->flags & SET_STYLE)
+			dstyle = dir_settings->display_style;
+
+		if (dir_settings->flags & SET_DETAILS)
+		{
+			dtype = dir_settings->details_type;
+			filer_window->view_type = dir_settings->view_type;
+		}
+
+		if (dir_settings->flags & SET_SORT)
+		{
+			s_type = dir_settings->sort_type;
+			s_order = dir_settings->sort_order;
+		}
+
+		if (dir_settings->flags & SET_THUMBS)
+			filer_window->show_thumbs = dir_settings->show_thumbs;
+		
+		if (dir_settings->flags & SET_FILTER)
+			filer_set_filter(filer_window,
+					   dir_settings->filter_type,
+					   dir_settings->filter);
+	}
+
 	/* Add all the user-interface elements & realise */
 	filer_add_widgets(filer_window, wm_class);
 	if (src_win)
 		gtk_window_set_position(GTK_WINDOW(filer_window->window),
 					GTK_WIN_POS_MOUSE);
 
+	if (dir_settings)
+	{
+		if (dir_settings->flags & SET_POSITION)
+		{
+			gtk_window_move(GTK_WINDOW(filer_window->window),
+					    dir_settings->x, dir_settings->y);
+		}
+		if (dir_settings->flags & SET_SIZE)
+		{
+			filer_window_set_size(filer_window,
+					      dir_settings->width,
+					      dir_settings->height);
+			force_resize = o_filer_auto_resize.int_value != RESIZE_NEVER;
+		}
+	}
+
 	/* Connect to all the signal handlers */
 	filer_add_signals(filer_window);
 
-	display_set_layout(filer_window, dstyle, dtype, TRUE);
+	display_set_layout(filer_window, dstyle, dtype, force_resize);
 	display_set_sort_type(filer_window, s_type, s_order);
 
 	/* Open the window after a timeout, or when scanning stops.
 	 * Do this before attaching, because attach() might tell us to
 	 * stop scanning (if a scan isn't needed).
 	 */
-	filer_window->open_timeout = gtk_timeout_add(500,
-					  (GtkFunction) open_filer_window,
+	filer_window->open_timeout = g_timeout_add(500,
+					  (GSourceFunc) open_filer_window,
 					  filer_window);
 
 	/* The view is created empty and then attach() is called, which
@@ -1376,6 +1631,8 @@ static void filer_add_widgets(FilerWindow *filer_window, const gchar *wm_class)
 	/* Create the top-level window widget */
 	filer_window->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 	filer_set_title(filer_window);
+	gtk_widget_set_name(filer_window->window, "rox-filer");
+
 	if (wm_class)
 		gtk_window_set_wmclass(GTK_WINDOW(filer_window->window),
 				       wm_class, PROJECT);
@@ -1462,9 +1719,9 @@ static void filer_add_signals(FilerWindow *filer_window)
 	GtkTargetEntry 	target_table[] =
 	{
 		{"text/uri-list", 0, TARGET_URI_LIST},
+		{"UTF8_STRING", 0, TARGET_STRING},
 		{"STRING", 0, TARGET_STRING},
 		{"COMPOUND_TEXT", 0, TARGET_STRING},/* XXX: Treats as STRING */
-		{"UTF8_STRING", 0, TARGET_STRING},
 	};
 
 	/* Events on the top-level window */
@@ -1513,7 +1770,7 @@ static void set_scanning_display(FilerWindow *filer_window, gboolean scanning)
 	if (scanning)
 		filer_set_title(filer_window);
 	else
-		gtk_timeout_add(300, (GtkFunction) clear_scanning_display,
+		g_timeout_add(300, (GSourceFunc) clear_scanning_display,
 				filer_window);
 }
 
@@ -1545,7 +1802,13 @@ void filer_update_all(void)
 		 */
 		next = next->next;
 
-		filer_update_dir(filer_window, TRUE);
+		/* Don't trigger a refresh if we're already scanning.
+		 * Otherwise, two views of a single directory will trigger
+		 * two scans.
+		 */
+		if (filer_window->directory &&
+		    !filer_window->directory->scanning)
+			filer_update_dir(filer_window, TRUE);
 	}
 }
 
@@ -1701,26 +1964,55 @@ void filer_set_title(FilerWindow *filer_window)
 {
 	gchar	*title = NULL;
 	guchar	*flags = "";
+	gchar  *hidden = "";
 
-	if (filer_window->scanning || filer_window->show_hidden ||
-				filer_window->show_thumbs)
+	if (filer_window->scanning ||
+	    filer_window->filter != FILER_SHOW_ALL ||
+	    filer_window->show_hidden || filer_window->show_thumbs)
 	{
 		if (o_short_flag_names.int_value)
 		{
+
+			switch(filer_window->filter) {
+			case FILER_SHOW_ALL:
+				hidden=filer_window->show_hidden? "A": "";
+				break;
+			case FILER_SHOW_GLOB:   hidden="G"; break;
+			case FILER_SHOW_REGEXP: hidden="R"; break;
+			default: break;
+			}
+
 			flags = g_strconcat(" +",
 				filer_window->scanning ? _("S") : "",
-				filer_window->show_hidden ? _("A") : "",
+				hidden,
 				filer_window->show_thumbs ? _("T") : "",
 				NULL);
 		}
 		else
 		{
+			switch(filer_window->filter) {
+			case FILER_SHOW_ALL:
+				hidden=g_strdup(filer_window->show_hidden? "All, ": "");
+				break;
+			case FILER_SHOW_GLOB:
+				hidden=g_strdup_printf("Glob (%s), ",
+						 filer_window->filter_string);
+				break;
+			case FILER_SHOW_REGEXP:
+				hidden=g_strdup_printf("Regexp (%s), ",
+						 filer_window->filter_string);
+				break;
+			default:
+				hidden=g_strdup("");
+				break;
+			}
 			flags = g_strconcat(" (",
 				filer_window->scanning ? _("Scanning, ") : "",
-				filer_window->show_hidden ? _("All, ") : "",
+				hidden,
 				filer_window->show_thumbs ? _("Thumbs, ") : "",
 				NULL);
 			flags[strlen(flags) - 2] = ')';
+			g_free(hidden);
 		}
 	}
 
@@ -1881,7 +2173,7 @@ static void filer_next_thumb(GObject *window, const gchar *path)
 	if (path)
 		dir_force_update_path(path);
 
-	gtk_idle_add((GtkFunction) filer_next_thumb_real, window);
+	g_idle_add((GSourceFunc) filer_next_thumb_real, window);
 }
 
 static void start_thumb_scanning(FilerWindow *filer_window)
@@ -2044,11 +2336,12 @@ static guchar *filer_create_uri_list(FilerWindow *filer_window)
 	view_get_iter(filer_window->view, &iter, VIEW_ITER_SELECTED);
 	while ((item = iter.next(&iter)))
 	{
-		gchar *uri, *path;
+		EscapedPath *uri;
+		char *path;
 		
 		path = g_strconcat(leader->str, item->leafname, NULL);
 		uri = encode_path_as_uri(path);
-		g_string_append(string, uri);
+		g_string_append(string, (char *) uri);
 		g_string_append(string, "\r\n");
 		g_free(path);
 		g_free(uri);
@@ -2319,7 +2612,7 @@ gint filer_motion_notify(FilerWindow *filer_window, GdkEventMotion *event)
 
 	if (view_count_selected(view) == 1)
 	{
-		if (!item->image)
+		if (item->base_type == TYPE_UNKNOWN)
 			item = dir_update_item(filer_window->directory,
 						item->leafname);
 
@@ -2331,7 +2624,7 @@ gint filer_motion_notify(FilerWindow *filer_window, GdkEventMotion *event)
 
 		drag_one_item(GTK_WIDGET(view), event,
 			make_path(filer_window->sym_path, item->leafname),
-			item, item->image);
+			item, di_image(item));
 #if 0
 		/* XXX: Use thumbnail */
 			item, view ? view->image : NULL);
@@ -2502,8 +2795,8 @@ void filer_set_autoscroll(FilerWindow *filer_window, gboolean auto_scroll)
 		if (filer_window->auto_scroll != -1)
 			return;		/* Already on! */
 
-		filer_window->auto_scroll = gtk_timeout_add(50,
-					(GtkFunction) as_timeout,
+		filer_window->auto_scroll = g_timeout_add(50,
+					(GSourceFunc) as_timeout,
 					filer_window);
 	}
 	else
@@ -2511,7 +2804,7 @@ void filer_set_autoscroll(FilerWindow *filer_window, gboolean auto_scroll)
 		if (filer_window->auto_scroll == -1)
 			return;		/* Already off! */
 
-		gtk_timeout_remove(filer_window->auto_scroll);
+		g_source_remove(filer_window->auto_scroll);
 		filer_window->auto_scroll = -1;
 	}
 }
@@ -2548,3 +2841,528 @@ void filer_refresh(FilerWindow *filer_window)
 	
 	full_refresh();
 }
+
+gboolean filer_match_filter(FilerWindow *filer_window, const gchar *filename)
+{
+	g_return_val_if_fail(filename != NULL, FALSE);
+
+	if(filename[0]=='.' &&
+	   (!filer_window->temp_show_hidden && !filer_window->show_hidden))
+		return FALSE;
+
+	switch(filer_window->filter) {
+	case FILER_SHOW_GLOB:
+		return fnmatch(filer_window->filter_string,
+			       filename, 0)==0;
+		
+	case FILER_SHOW_REGEXP: /* Unimplemented */
+
+	case FILER_SHOW_ALL:
+	default:
+		break;
+	}
+	return TRUE;
+}
+
+/* Provided to hide the implementation */
+void filer_set_hidden(FilerWindow *filer_window, gboolean hidden)
+{
+	filer_window->show_hidden=hidden;
+}
+
+/* Set the filter type. Returns TRUE if the type has changed
+ * (must call filer_detach_rescan).
+ */
+gboolean filer_set_filter(FilerWindow *filer_window, FilterType type,
+			     const gchar *filter_string)
+{
+	/* Is this new filter the same as the old one? */
+	if (filer_window->filter == type)
+	{
+		switch(filer_window->filter)
+		{
+		case FILER_SHOW_ALL:
+			return FALSE;
+		case FILER_SHOW_GLOB:
+		case FILER_SHOW_REGEXP:
+			if (strcmp(filer_window->filter_string,
+				   filter_string) == 0)
+				return FALSE;
+			break;
+		}
+	}
+
+	/* Clean up old filter */
+	if (filer_window->filter_string)
+	{
+		g_free(filer_window->filter_string);
+		filer_window->filter_string = NULL;
+	}
+	/* Also clean up compiled regexp when implemented */
+
+	filer_window->filter = type;
+
+	switch(type)
+	{
+	case FILER_SHOW_ALL:
+		/* No extra work */
+		break;
+
+	case FILER_SHOW_GLOB:
+		filer_window->filter_string = g_strdup(filter_string);
+		break;
+
+	case FILER_SHOW_REGEXP:
+		filer_window->filter_string = g_strdup(filter_string);
+		/* Compile the pattern */
+		break;
+
+	default:
+		/* oops */
+		filer_window->filter = FILER_SHOW_ALL;
+		g_warning("Impossible: filter type %d", type);
+		break;
+	}
+
+	return TRUE;
+}
+
+/* Setting stuff */
+static Settings *settings_new(const char *path)
+{
+	Settings *set;
+
+	set=g_new(Settings, 1);
+	memset(set, 0, sizeof(Settings));
+	if(path)
+		set->path=g_strdup(path);
+
+	return set;
+}
+
+static void settings_free(Settings *set)
+{
+	g_free(set->path);
+	if(set->filter)
+		g_free(set->filter);
+	g_free(set);
+}
+
+static gboolean free_settings(gpointer key, gpointer value, gpointer data)
+{
+	if(!data || strcmp(data, key)==0)
+		settings_free(value);
+
+	return TRUE;
+}
+
+static void store_settings(Settings *set)
+{
+	Settings *old;
+
+	old=g_hash_table_lookup(settings_table, set->path);
+	if(old) {
+		g_hash_table_foreach_remove(settings_table, free_settings,
+					    set->path);
+	}
+
+	g_hash_table_insert(settings_table, set->path, set);
+}
+
+/* TODO: use symbolic names in the XML file where possible */
+static void load_from_node(Settings *set, xmlDocPtr doc, xmlNodePtr node)
+{
+	xmlChar *str=NULL;
+	
+	str=xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
+	
+	if(strcmp(node->name, "X") == 0) {
+		set->x=atoi(str);
+		set->flags|=SET_POSITION;		
+	} else if(strcmp(node->name, "Y") == 0) {
+		set->y=atoi(str);
+		set->flags|=SET_POSITION;		
+	} else if(strcmp(node->name, "Width") == 0) {
+		set->width=atoi(str);
+		set->flags|=SET_SIZE;		
+	} else if(strcmp(node->name, "Height") == 0) {
+		set->height=atoi(str);
+		set->flags|=SET_SIZE;		
+	} else if(strcmp(node->name, "ShowHidden") == 0) {
+		set->show_hidden=atoi(str);
+		set->flags|=SET_HIDDEN;		
+	} else if(strcmp(node->name, "ViewType") == 0) {
+		set->view_type=atoi(str);
+		set->flags|=SET_DETAILS;		
+	} else if(strcmp(node->name, "DetailsType") == 0) {
+		set->details_type=atoi(str);
+		set->flags|=SET_DETAILS;		
+	} else if(strcmp(node->name, "SortType") == 0) {
+		set->sort_type=atoi(str);
+		set->flags|=SET_SORT;		
+	} else if(strcmp(node->name, "SortOrder") == 0) {
+		set->sort_order=atoi(str);
+		set->flags|=SET_SORT;		
+	} else if(strcmp(node->name, "DisplayStyle") == 0) {
+		set->display_style=atoi(str);
+		set->flags|=SET_STYLE;		
+	} else if(strcmp(node->name, "ShowThumbs") == 0) {
+		set->show_thumbs=atoi(str);
+		set->flags|=SET_THUMBS;		
+	} else if(strcmp(node->name, "FilterType") == 0) {
+		set->filter_type=atoi(str);
+		set->flags|=SET_FILTER;		
+	} else if(strcmp(node->name, "Filter") == 0) {
+		set->filter=g_strdup(str);
+		set->flags|=SET_FILTER;		
+	}
+	
+	if(str)
+		xmlFree(str);
+}
+
+static void load_settings(void)
+{
+	gchar *path;
+	XMLwrapper *settings_doc=NULL;
+
+	path=choices_find_path_load("Settings.xml", "ROX-Filer");
+	if(path) {
+		settings_doc=xml_new(path);
+		g_free(path);
+	}
+
+	if(!settings_table)
+		settings_table=g_hash_table_new(g_str_hash, g_str_equal);
+
+	if(settings_doc) {
+		xmlNodePtr node, subnode;
+
+		node = xmlDocGetRootElement(settings_doc->doc);
+
+		for (node = node->xmlChildrenNode; node; node = node->next)
+		{
+			Settings *set;
+			xmlChar *path;
+			
+			if (node->type != XML_ELEMENT_NODE)
+				continue;
+			if (strcmp(node->name, "FilerWindow") != 0)
+				continue;
+
+			path=xmlGetProp(node, "path");
+			set=settings_new(path);
+			xmlFree(path);
+
+			for (subnode=node->xmlChildrenNode; subnode;
+			     subnode=subnode->next) {
+				
+				if (subnode->type != XML_ELEMENT_NODE)
+					continue;
+				load_from_node(set, settings_doc->doc,
+					       subnode);
+			}
+
+			store_settings(set);
+		}
+		g_object_unref(settings_doc);
+	}
+}
+
+static void add_nodes(gpointer key, gpointer value, gpointer data)
+{
+	xmlNodePtr node=(xmlNodePtr) data;
+	xmlNodePtr sub;
+	Settings *set=(Settings *) value;
+	char *tmp;
+
+	sub=xmlNewChild(node, NULL, "FilerWindow", NULL);
+
+	xmlSetProp(sub, "path", set->path);
+	
+	if(set->flags & SET_POSITION) {
+		tmp=g_strdup_printf("%d", set->x);
+		xmlNewChild(sub, NULL, "X", tmp);
+		g_free(tmp);
+		tmp=g_strdup_printf("%d", set->y);
+		xmlNewChild(sub, NULL, "Y", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_SIZE) {
+		tmp=g_strdup_printf("%d", set->width);
+		xmlNewChild(sub, NULL, "Width", tmp);
+		g_free(tmp);
+		tmp=g_strdup_printf("%d", set->height);
+		xmlNewChild(sub, NULL, "Height", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_HIDDEN) {
+		tmp=g_strdup_printf("%d", set->show_hidden);
+		xmlNewChild(sub, NULL, "ShowHidden", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_STYLE) {
+		tmp=g_strdup_printf("%d", set->display_style);
+		xmlNewChild(sub, NULL, "DisplayStyle", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_SORT) {
+		tmp=g_strdup_printf("%d", set->sort_type);
+		xmlNewChild(sub, NULL, "SortType", tmp);
+		g_free(tmp);
+		tmp=g_strdup_printf("%d", set->sort_order);
+		xmlNewChild(sub, NULL, "SortOrder", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_DETAILS) {
+		tmp=g_strdup_printf("%d", set->view_type);
+		xmlNewChild(sub, NULL, "ViewType", tmp);
+		g_free(tmp);
+		tmp=g_strdup_printf("%d", set->details_type);
+		xmlNewChild(sub, NULL, "DetailsType", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_STYLE) {
+		tmp=g_strdup_printf("%d", set->show_thumbs);
+		xmlNewChild(sub, NULL, "ShowThumbs", tmp);
+		g_free(tmp);
+	}
+	if(set->flags & SET_FILTER) {
+		tmp=g_strdup_printf("%d", set->filter_type);
+		xmlNewChild(sub, NULL, "FilterType", tmp);
+		g_free(tmp);
+		if(set->filter && set->filter[0])
+			xmlNewChild(sub, NULL, "Filter", set->filter);
+	}
+}
+
+static void save_settings(void)
+{
+	gchar *path;
+
+	path=choices_find_path_save("Settings.xml", "ROX-Filer", TRUE);
+	if(path) {
+		xmlDocPtr doc = xmlNewDoc("1.0");
+		xmlDocSetRootElement(doc, xmlNewDocNode(doc, NULL,
+							"Settings", NULL));
+		
+		g_hash_table_foreach(settings_table, add_nodes,
+					    xmlDocGetRootElement(doc));
+		
+		save_xml_file(doc, path);
+		
+		g_free(path);
+		xmlFreeDoc(doc);
+	}
+
+}
+
+static void check_settings(FilerWindow *filer_window)
+{
+	Settings *set;
+
+	set=(Settings *) g_hash_table_lookup(settings_table,
+					      filer_window->sym_path);
+
+	if(set) {
+		if(set->flags & SET_POSITION) 
+			gtk_window_move(GTK_WINDOW(filer_window->window),
+					    set->x, set->y);
+		if(set->flags & SET_SIZE) 
+			filer_window_set_size(filer_window, set->width,
+					      set->height);
+		if(set->flags & SET_HIDDEN)
+			filer_set_hidden(filer_window, set->show_hidden);
+
+		if(set->flags & (SET_STYLE|SET_DETAILS)) {
+			DisplayStyle style=filer_window->display_style;
+			DetailsType  details=filer_window->details_type;
+
+			if(set->flags & SET_STYLE)
+				style=set->display_style;
+
+			if(set->flags & SET_DETAILS) {
+				details=set->details_type;
+
+				filer_set_view_type(filer_window,
+						    set->view_type);
+			}
+
+			display_set_layout(filer_window, style,
+					   details, FALSE);
+		}
+
+		if(set->flags & SET_SORT)
+			display_set_sort_type(filer_window,
+					      set->sort_type,
+					      set->sort_order);
+
+		if(set->flags & SET_THUMBS)
+			display_set_thumbs(filer_window,
+					   set->show_thumbs);
+		
+		if(set->flags & SET_FILTER)
+			display_set_filter(filer_window,
+					   set->filter_type,
+					   set->filter);
+	}
+
+}
+
+typedef struct settings_window {
+	GtkWidget *window;
+
+	GtkWidget *pos, *size, *hidden, *style, *sort, *details,
+		*thumbs, *filter;
+
+	Settings *set;
+} SettingsWindow;
+
+
+static void settings_response(GtkWidget *window, gint response,
+			      SettingsWindow *set_win)
+{
+	if(response==GTK_RESPONSE_OK) {
+		gint flags=0;
+
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->pos)))
+			flags|=SET_POSITION;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->size)))
+			flags|=SET_SIZE;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->hidden)))
+			flags|=SET_HIDDEN;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->style)))
+			flags|=SET_STYLE;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->sort)))
+			flags|=SET_SORT;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->details)))
+			flags|=SET_DETAILS;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->thumbs)))
+			flags|=SET_THUMBS;
+		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(set_win->filter)))
+			flags|=SET_FILTER;
+
+		set_win->set->flags=flags;
+		store_settings(set_win->set);
+		save_settings();
+	}
+	
+	gtk_widget_destroy(window);
+}
+
+void filer_save_settings(FilerWindow *fwin)
+{
+	SettingsWindow *set_win;
+	GtkWidget *vbox;
+	GtkWidget *path;
+	gint x, y;
+	
+	Settings *set=settings_new(fwin->sym_path);
+
+	gtk_window_get_position(GTK_WINDOW(fwin->window),&x, &y);
+	set->flags|=SET_POSITION;
+	set->x=x;
+	set->y=y;
+
+	gtk_window_get_size(GTK_WINDOW(fwin->window),&x, &y);
+	set->flags|=SET_SIZE;
+	set->width=x;
+	set->height=y;
+
+	set->flags|=SET_HIDDEN;
+	set->show_hidden=fwin->show_hidden;
+
+	set->flags|=SET_STYLE;
+	set->display_style=fwin->display_style;
+
+	set->flags|=SET_SORT;
+	set->sort_type=fwin->sort_type;
+	set->sort_order=fwin->sort_order;
+
+	set->flags|=SET_DETAILS;
+	set->view_type=fwin->view_type;
+	set->details_type=fwin->details_type;
+
+	set->flags|=SET_THUMBS;
+	set->show_thumbs=fwin->show_thumbs;
+
+	set->flags|=SET_FILTER;
+	set->filter_type=fwin->filter;
+	if(fwin->filter_string)
+		set->filter=g_strdup(fwin->filter_string);
+
+	/* Store other parameters
+	*/
+	set_win=g_new(SettingsWindow, 1);
+
+	set_win->window=gtk_dialog_new();
+	number_of_windows++;
+	
+	gtk_dialog_add_button(GTK_DIALOG(set_win->window),
+			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
+	gtk_dialog_add_button(GTK_DIALOG(set_win->window),
+			GTK_STOCK_OK, GTK_RESPONSE_OK);
+
+	g_signal_connect(set_win->window, "destroy",
+			 G_CALLBACK(one_less_window), NULL);
+	g_signal_connect_swapped(set_win->window, "destroy",
+			 G_CALLBACK(g_free), set_win);
+
+	gtk_window_set_title(GTK_WINDOW(set_win->window),
+			     _("Select display properties to save"));
+
+	vbox=GTK_DIALOG(set_win->window)->vbox;
+
+	path=gtk_label_new(set->path);
+	gtk_box_pack_start(GTK_BOX(vbox), path, FALSE, FALSE, 2);
+
+	set_win->pos=gtk_check_button_new_with_label(_("Position"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->pos, FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->pos),
+				     set->flags & SET_POSITION);
+	
+	set_win->size=gtk_check_button_new_with_label(_("Size"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->size, FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->size),
+				     set->flags & SET_SIZE);
+	
+	set_win->hidden=gtk_check_button_new_with_label(_("Show hidden"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->hidden,
+			   FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->hidden),
+				     set->flags & SET_HIDDEN);
+
+	set_win->style=gtk_check_button_new_with_label(_("Display style"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->style,
+			   FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->style),
+				     set->flags & SET_STYLE);
+
+	set_win->sort=gtk_check_button_new_with_label(_("Sort type and order"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->sort, FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->sort),
+				     set->flags & SET_SORT);
+	
+	set_win->details=gtk_check_button_new_with_label(_("Details"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->details, FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->details),
+				     set->flags & SET_DETAILS);
+	
+	set_win->thumbs=gtk_check_button_new_with_label(_("Thumbnails"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->thumbs,
+			   FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->thumbs),
+				     set->flags & SET_THUMBS);
+
+	set_win->filter=gtk_check_button_new_with_label(_("Filter"));
+	gtk_box_pack_start(GTK_BOX(vbox), set_win->filter,
+			   FALSE, FALSE, 2);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_win->filter),
+				     set->flags & SET_FILTER);
+
+	set_win->set=set;
+	g_signal_connect(set_win->window, "response",
+			 G_CALLBACK(settings_response), set_win);
+
+	gtk_widget_show_all(set_win->window);
+}
