@@ -1,8 +1,8 @@
 /*
- * $Id: gui_support.c,v 1.111 2003/10/07 14:01:55 tal197 Exp $
+ * $Id: gui_support.c,v 1.127 2005/01/16 16:48:01 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -46,9 +46,13 @@
 #include "choices.h"
 #include "options.h"
 
-/* XXX: RandR will break this! */
 gint	screen_width, screen_height;
 
+gint		n_monitors;
+GdkRectangle	*monitor_geom = NULL;
+gint		monitor_width, monitor_height;
+MonitorAdjacent *monitor_adjacent;
+
 static GdkAtom xa_cardinal;
 
 static GtkWidget *current_dialog = NULL;
@@ -61,6 +65,51 @@ static gint tip_timeout = 0;	/* When primed */
 static void run_error_info_dialog(GtkMessageType type, const char *message,
 				  va_list args);
 static GType simple_image_get_type(void);
+static void gui_get_monitor_adjacent(int monitor, MonitorAdjacent *adj);
+
+void gui_store_screen_geometry(GdkScreen *screen)
+{
+	gint mon;
+
+	screen_width = gdk_screen_get_width(screen);
+	screen_height = gdk_screen_get_height(screen);
+
+	if (monitor_adjacent)
+		g_free(monitor_adjacent);
+
+	monitor_width = monitor_height = G_MAXINT;
+	n_monitors = gdk_screen_get_n_monitors(screen);
+	if (monitor_geom)
+		g_free(monitor_geom);
+	monitor_geom = g_new(GdkRectangle, n_monitors ? n_monitors : 1);
+
+	if (n_monitors)
+	{
+		for (mon = 0; mon < n_monitors; ++mon)
+		{
+			gdk_screen_get_monitor_geometry(screen, mon,
+					&monitor_geom[mon]);
+			if (monitor_geom[mon].width < monitor_width)
+				monitor_width = monitor_geom[mon].width; 
+			if (monitor_geom[mon].height < monitor_height)
+				monitor_height = monitor_geom[mon].height; 
+		}
+		monitor_adjacent = g_new(MonitorAdjacent, n_monitors);
+		for (mon = 0; mon < n_monitors; ++mon)
+		{
+			gui_get_monitor_adjacent(mon, &monitor_adjacent[mon]);
+		}
+	}
+	else
+	{
+		n_monitors = 1;
+		monitor_geom[0].x = monitor_geom[0].y = 0;
+		monitor_width = monitor_geom[0].width = screen_width;
+		monitor_height = monitor_geom[0].height = screen_height;
+		monitor_adjacent = g_new0(MonitorAdjacent, 1);
+	}
+
+}
 
 void gui_support_init()
 {
@@ -68,11 +117,7 @@ void gui_support_init()
 	
 	xa_cardinal = gdk_atom_intern("CARDINAL", FALSE);
 
-	/* This call starts returning strange values after a while, so get
-	 * the result here during init.
-	 */
-	gdk_drawable_get_size(gdk_get_default_root_window(),
-			    &screen_width, &screen_height);
+	gui_store_screen_geometry(gdk_screen_get_default());
 
 	/* Work around the scrollbar placement bug */
 	klass = g_type_class_ref(gtk_scrolled_window_get_type());
@@ -181,9 +226,8 @@ void set_cardinal_property(GdkWindow *window, GdkAtom prop, guint32 value)
 void make_panel_window(GtkWidget *widget)
 {
 	static gboolean need_init = TRUE;
-	static GdkAtom xa_state, xa_atom, xa_net_state, xa_hints, xa_win_hints;
+	static GdkAtom xa_state, xa_atom, xa_hints, xa_win_hints;
 	static GdkAtom xa_NET_WM_DESKTOP;
-	static GdkAtom state_list[3];
 	GdkWindow *window = widget->window;
 	gint32  wm_hints_values[] = {1, False, 0, 0, 0, 0, 0, 0};
 	GdkAtom	wm_protocols[2];
@@ -201,19 +245,9 @@ void make_panel_window(GtkWidget *widget)
 		xa_win_hints = gdk_atom_intern("_WIN_HINTS", FALSE);
 		xa_state = gdk_atom_intern("_WIN_STATE", FALSE);
 		xa_atom = gdk_atom_intern("ATOM", FALSE);
-		xa_net_state = gdk_atom_intern("_NET_WM_STATE", FALSE);
 		xa_hints = gdk_atom_intern("WM_HINTS", FALSE);
 		xa_NET_WM_DESKTOP = gdk_atom_intern("_NET_WM_DESKTOP", FALSE);
 
-		/* Note: Starting with Gtk+-1.3.12, Gtk+ converts GdkAtoms
-		 * to X atoms automatically when the type is ATOM.
-		 */
-		state_list[0] = gdk_atom_intern("_NET_WM_STATE_STICKY", FALSE);
-		state_list[1] = gdk_atom_intern("_NET_WM_STATE_SKIP_PAGER",
-						FALSE);
-		state_list[2] = gdk_atom_intern("_NET_WM_STATE_SKIP_TASKBAR",
-						FALSE);
-		
 		need_init = FALSE;
 	}
 	
@@ -221,10 +255,6 @@ void make_panel_window(GtkWidget *widget)
 	gdk_window_set_functions(window, 0);
 	gtk_window_set_resizable(GTK_WINDOW(widget), FALSE);
 
-	/* Note: DON'T do gtk_window_stick(). Setting the state via
-	 * gdk will override our other atoms (pager/taskbar).
-	 */
-
 	/* Don't hide panel/pinboard windows initially (WIN_STATE_HIDDEN).
 	 * Needed for IceWM - Christopher Arndt <chris.arndt@web.de>
 	 */
@@ -239,9 +269,6 @@ void make_panel_window(GtkWidget *widget)
 	/* Appear on all workspaces */
 	set_cardinal_property(window, xa_NET_WM_DESKTOP, 0xffffffff);
 
-	gdk_property_change(window, xa_net_state, xa_atom, 32,
-			GDK_PROP_MODE_APPEND, (guchar *) state_list, 3);
-
 	gdk_property_change(window, xa_hints, xa_hints, 32,
 			GDK_PROP_MODE_REPLACE, (guchar *) wm_hints_values,
 			sizeof(wm_hints_values) / sizeof(gint32));
@@ -252,6 +279,20 @@ void make_panel_window(GtkWidget *widget)
 			gdk_atom_intern("WM_PROTOCOLS", FALSE), xa_atom, 32,
 			GDK_PROP_MODE_REPLACE, (guchar *) wm_protocols,
 			sizeof(wm_protocols) / sizeof(GdkAtom));
+
+	gdk_window_set_skip_taskbar_hint(window, TRUE);
+	gdk_window_set_skip_pager_hint(window, TRUE);
+
+	if (g_object_class_find_property(G_OBJECT_GET_CLASS(widget),
+					"accept_focus"))
+	{
+		GValue vfalse = { 0, };
+		g_value_init(&vfalse, G_TYPE_BOOLEAN);
+		g_value_set_boolean(&vfalse, FALSE);
+		g_object_set_property(G_OBJECT(widget),
+					"accept_focus", &vfalse);
+		g_value_unset(&vfalse);
+	}
 }
 
 static gboolean error_idle_cb(gpointer data)
@@ -295,7 +336,7 @@ void delayed_error(const char *error, ...)
 	else
 	{
 		delayed_error_data = new;
-		gtk_idle_add(error_idle_cb, &delayed_error_data);
+		g_idle_add(error_idle_cb, &delayed_error_data);
 
 		number_of_windows++;
 	}
@@ -327,7 +368,7 @@ GtkWidget *new_help_button(HelpFunc show_help, gpointer data)
 	
 	b = gtk_button_new();
 	gtk_button_set_relief(GTK_BUTTON(b), GTK_RELIEF_NONE);
-	icon = gtk_image_new_from_stock(GTK_STOCK_DIALOG_INFO,
+	icon = gtk_image_new_from_stock(GTK_STOCK_HELP,
 					GTK_ICON_SIZE_SMALL_TOOLBAR);
 	gtk_container_add(GTK_CONTAINER(b), icon);
 	g_signal_connect_swapped(b, "clicked", G_CALLBACK(show_help), data);
@@ -411,17 +452,24 @@ gboolean get_pointer_xy(int *x, int *y)
 void centre_window(GdkWindow *window, int x, int y)
 {
 	int	w, h;
+	int m;
 
 	g_return_if_fail(window != NULL);
 
+	m = gdk_screen_get_monitor_at_point(gdk_screen_get_default(), x, y);
+
 	gdk_drawable_get_size(window, &w, &h);
 	
 	x -= w / 2;
 	y -= h / 2;
 
 	gdk_window_move(window,
-		CLAMP(x, DECOR_BORDER, screen_width - w - DECOR_BORDER),
-		CLAMP(y, DECOR_BORDER, screen_height - h - DECOR_BORDER));
+		CLAMP(x, DECOR_BORDER + monitor_geom[m].x,
+			monitor_geom[m].x + monitor_geom[m].width 
+			- w - DECOR_BORDER),
+		CLAMP(y, DECOR_BORDER + monitor_geom[m].y,
+			monitor_geom[m].y + monitor_geom[m].height 
+			- h - DECOR_BORDER));
 }
 
 static void run_error_info_dialog(GtkMessageType type, const char *message,
@@ -465,14 +513,14 @@ static gboolean end_wink(gpointer data)
 
 static void cancel_wink(void)
 {
-	gtk_timeout_remove(wink_timeout);
+	g_source_remove(wink_timeout);
 	end_wink(NULL);
 }
 
 static void wink_widget_died(gpointer data)
 {
 	current_wink_widget = NULL;
-	gtk_timeout_remove(wink_timeout);
+	g_source_remove(wink_timeout);
 }
 
 /* Draw a black box around this widget, briefly.
@@ -488,7 +536,7 @@ void wink_widget(GtkWidget *widget)
 	current_wink_widget = widget;
 	gtk_drag_highlight(current_wink_widget);
 	
-	wink_timeout = gtk_timeout_add(300, (GtkFunction) end_wink, NULL);
+	wink_timeout = g_timeout_add(300, (GSourceFunc) end_wink, NULL);
 
 	wink_destroy = g_signal_connect_swapped(widget, "destroy",
 				G_CALLBACK(wink_widget_died), NULL);
@@ -505,7 +553,7 @@ static gboolean idle_destroy_cb(GtkWidget *widget)
 void destroy_on_idle(GtkWidget *widget)
 {
 	gtk_widget_ref(widget);
-	gtk_idle_add((GtkFunction) idle_destroy_cb, widget);
+	g_idle_add((GSourceFunc) idle_destroy_cb, widget);
 }
 
 /* Spawn a child process (as spawn_full), and report errors.
@@ -688,10 +736,11 @@ void tooltip_show(guchar *text)
 	GtkWidget *label;
 	int	x, y, py;
 	int	w, h;
+	int m;
 
 	if (tip_timeout)
 	{
-		gtk_timeout_remove(tip_timeout);
+		g_source_remove(tip_timeout);
 		tip_timeout = 0;
 	}
 
@@ -722,16 +771,20 @@ void tooltip_show(guchar *text)
 	h = tip_widget->allocation.height;
 	gdk_window_get_pointer(NULL, &x, &py, NULL);
 
+	m = gdk_screen_get_monitor_at_point(gdk_screen_get_default(), x, py);
+	
 	x -= w / 2;
 	y = py + 12; /* I don't know the pointer height so I use a constant */
 
 	/* Now check for screen boundaries */
-	x = CLAMP(x, 0, screen_width - w);
-	y = CLAMP(y, 0, screen_height - h);
+	x = CLAMP(x, monitor_geom[m].x,
+			monitor_geom[m].x + monitor_geom[m].width - w);
+	y = CLAMP(y, monitor_geom[m].y,
+			monitor_geom[m].y + monitor_geom[m].height - h);
 
 	/* And again test if pointer is over the tooltip window */
 	if (py >= y && py <= y + h)
-		y = py - h- 2;
+		y = py - h - 2;
 	gtk_window_move(GTK_WINDOW(tip_widget), x, y);
 	gtk_widget_show(tip_widget);
 
@@ -754,11 +807,11 @@ void tooltip_prime(GtkFunction callback, GObject *object)
 	delay = now - tip_time > 2 ? 1000 : 200;
 
 	g_object_ref(object);
-	tip_timeout = gtk_timeout_add_full(delay,
-					   (GtkFunction) callback,
-					   NULL,
-					   object,
-					   g_object_unref);
+	tip_timeout = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE,
+					 delay,
+					 (GSourceFunc) callback,
+					 object,
+					 g_object_unref);
 }
 
 /* Like gtk_widget_modify_font, but copes with font_desc == NULL */
@@ -928,7 +981,8 @@ gint radios_get_value(Radios *radios)
 	return -1;
 }
 
-/* Convert a list of URIs as a string into a GList of URIs.
+/* Convert a list of URIs as a string into a GList of EscapedPath URIs.
+ * No unescaping is done.
  * Lines beginning with # are skipped.
  * The text block passed in is zero terminated (after the final CRLF)
  */
@@ -939,7 +993,6 @@ GList *uri_list_to_glist(const char *uri_list)
 	while (*uri_list)
 	{
 		char	*linebreak;
-		char	*uri;
 		int	length;
 
 		linebreak = strchr(uri_list, 13);
@@ -955,13 +1008,7 @@ GList *uri_list_to_glist(const char *uri_list)
 		length = linebreak - uri_list;
 
 		if (length && uri_list[0] != '#')
-		{
-			char *tmp;
-			tmp = g_strndup(uri_list, length);
-			uri = unescape_uri(tmp);
-			g_free(tmp);
-			list = g_list_append(list, uri);
-		}
+			list = g_list_append(list, g_strndup(uri_list, length));
 
 		uri_list = linebreak + 2;
 	}
@@ -1007,25 +1054,9 @@ static void simple_image_size_request(GtkWidget      *widget,
 void render_pixbuf(GdkPixbuf *pixbuf, GdkDrawable *target, GdkGC *gc,
 		   int x, int y, int width, int height)
 {
-#if GTK_MAJOR_VERSION == 2 && GTK_MINOR_VERSION > 1
 	gdk_draw_pixbuf(target, gc, pixbuf, 0, 0, x, y, width, height,
 		        GDK_RGB_DITHER_NORMAL, 0, 0);
 
-#else
-	static gboolean warned = FALSE;
-	if (gtk_minor_version == 0)
-		GDK_DRAWABLE_GET_CLASS(target)->_draw_pixbuf(target, gc, pixbuf,
-				   0, 0, x, y, width, height,
-				   GDK_RGB_DITHER_NORMAL, 0, 0);
-	else if (!warned)
-	{
-		delayed_error(_("Pinboard icons cannot be drawn because "
-				"ROX-Filer was compiled against GTK+-2.0 "
-				"but is running with GTK+-2.2. Please "
-				"recompile it."));
-		warned = TRUE;
-	}
-#endif
 }
 
 static gint simple_image_expose(GtkWidget *widget, GdkEventExpose *event)
@@ -1103,3 +1134,257 @@ GtkWidget *simple_image_new(GdkPixbuf *pixbuf)
 
 	return GTK_WIDGET(image);
 }
+
+/* Whether a line l1 long starting from n1 overlaps a line l2 from n2 */
+inline static gboolean gui_ranges_overlap(int n1, int l1, int n2, int l2)
+{
+	return (n1 > n2 && n1 < n2 + l2) ||
+		(n1 + l1 > n2 && n1 + l1 < n2 + l2) ||
+		(n1 <= n2 && n1 + l1 >= n2 + l2);
+}
+
+static void gui_get_monitor_adjacent(int monitor, MonitorAdjacent *adj)
+{
+	int m;
+
+	adj->left = adj->right = adj->top = adj->bottom = FALSE;
+
+	for (m = 0; m < n_monitors; ++m)
+	{
+		if (m == monitor)
+			continue;
+		if (gui_ranges_overlap(monitor_geom[m].y,
+				monitor_geom[m].height,
+				monitor_geom[monitor].y,
+				monitor_geom[monitor].height))
+		{
+			if (monitor_geom[m].x < monitor_geom[monitor].x)
+			{
+				adj->left = TRUE;
+			}
+			else if (monitor_geom[m].x > monitor_geom[monitor].x)
+			{
+				adj->right = TRUE;
+			}
+		}
+		if (gui_ranges_overlap(monitor_geom[m].x,
+				monitor_geom[m].width,
+				monitor_geom[monitor].x,
+				monitor_geom[monitor].width))
+		{
+			if (monitor_geom[m].y < monitor_geom[monitor].y)
+			{
+				adj->top = TRUE;
+			}
+			else if (monitor_geom[m].y > monitor_geom[monitor].y)
+			{
+				adj->bottom = TRUE;
+			}
+		}
+	}
+}
+
+static void rox_wmspec_change_state(gboolean add, GdkWindow *window,
+				    GdkAtom state1, GdkAtom state2)
+{
+	GdkDisplay *display = gdk_drawable_get_display(GDK_DRAWABLE(window));
+	XEvent xev;
+	
+#define _NET_WM_STATE_REMOVE        0    /* remove/unset property */
+#define _NET_WM_STATE_ADD           1    /* add/set property */
+#define _NET_WM_STATE_TOGGLE        2    /* toggle property  */  
+
+	xev.xclient.type = ClientMessage;
+	xev.xclient.serial = 0;
+	xev.xclient.send_event = True;
+	xev.xclient.window = GDK_WINDOW_XID(window);
+	xev.xclient.message_type = gdk_x11_get_xatom_by_name_for_display(
+			display, "_NET_WM_STATE");
+	xev.xclient.format = 32;
+	xev.xclient.data.l[0] = add ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE;
+	xev.xclient.data.l[1] = gdk_x11_atom_to_xatom_for_display(display,
+			state1);
+	xev.xclient.data.l[2] = gdk_x11_atom_to_xatom_for_display(display,
+			state2);
+	xev.xclient.data.l[3] = 0;
+	xev.xclient.data.l[4] = 0;
+	
+	XSendEvent(GDK_DISPLAY_XDISPLAY(display),
+		   GDK_WINDOW_XID(
+			gdk_screen_get_root_window(
+				gdk_drawable_get_screen(GDK_DRAWABLE(window)))),
+		   False,
+		   SubstructureRedirectMask | SubstructureNotifyMask,
+		   &xev);
+}
+
+void keep_below(GdkWindow *window, gboolean setting)
+{
+	g_return_if_fail(GDK_IS_WINDOW(window));
+
+	if (GDK_WINDOW_DESTROYED(window))
+		return;
+
+	if (gdk_window_is_visible(window))
+	{
+		if (setting)
+		{
+			rox_wmspec_change_state(FALSE, window,
+				gdk_atom_intern("_NET_WM_STATE_ABOVE", FALSE),
+				GDK_NONE);
+		}
+		rox_wmspec_change_state(setting, window,
+				gdk_atom_intern("_NET_WM_STATE_BELOW", FALSE),
+				GDK_NONE);
+	}
+#if 0
+	else
+	{
+#if GTK_CHECK_VERSION(2,4,0)
+	  gdk_synthesize_window_state(window,
+				setting ? GDK_WINDOW_STATE_ABOVE :
+					GDK_WINDOW_STATE_BELOW,
+				setting ? GDK_WINDOW_STATE_BELOW : 0);
+#endif
+	}
+#endif
+}
+
+static void
+size_prepared_cb (GdkPixbufLoader *loader, 
+		  int              width,
+		  int              height,
+		  gpointer         data)
+{
+	struct {
+		gint width;
+		gint height;
+		gboolean preserve_aspect_ratio;
+	} *info = data;
+
+	g_return_if_fail (width > 0 && height > 0);
+
+	if(info->preserve_aspect_ratio) {
+		if ((double)height * (double)info->width >
+		    (double)width * (double)info->height) {
+			width = 0.5 + (double)width * (double)info->height / (double)height;
+			height = info->height;
+		} else {
+			height = 0.5 + (double)height * (double)info->width / (double)width;
+			width = info->width;
+		}
+	} else {
+		width = info->width;
+		height = info->height;
+	}
+	
+	gdk_pixbuf_loader_set_size (loader, width, height);
+}
+
+/**
+ * rox_pixbuf_new_from_file_at_scale:
+ * @filename: Name of file to load.
+ * @width: The width the image should have
+ * @height: The height the image should have
+ * @preserve_aspect_ratio: %TRUE to preserve the image's aspect ratio
+ * @error: Return location for an error
+ *
+ * Creates a new pixbuf by loading an image from a file.  The file format is
+ * detected automatically. If %NULL is returned, then @error will be set.
+ * Possible errors are in the #GDK_PIXBUF_ERROR and #G_FILE_ERROR domains.
+ * The image will be scaled to fit in the requested size, optionally preserving
+ * the image's aspect ratio.
+ *
+ * Return value: A newly-created pixbuf with a reference count of 1, or %NULL 
+ * if any of several error conditions occurred:  the file could not be opened,
+ * there was no loader for the file's format, there was not enough memory to
+ * allocate the image buffer, or the image file contained invalid data.
+ *
+ * Taken from GTK 2.6.
+ **/
+GdkPixbuf *
+rox_pixbuf_new_from_file_at_scale (const char *filename,
+				   int         width, 
+				   int         height,
+				   gboolean    preserve_aspect_ratio,
+				   GError    **error)
+{
+
+	GdkPixbufLoader *loader;
+	GdkPixbuf       *pixbuf;
+
+	guchar buffer [4096];
+	int length;
+	FILE *f;
+	struct {
+		gint width;
+		gint height;
+		gboolean preserve_aspect_ratio;
+	} info;
+
+	g_return_val_if_fail (filename != NULL, NULL);
+        g_return_val_if_fail (width > 0 && height > 0, NULL);
+
+	f = fopen (filename, "rb");
+	if (!f) {
+                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
+                                                           NULL, NULL, NULL);
+                g_set_error (error,
+                             G_FILE_ERROR,
+                             g_file_error_from_errno (errno),
+                             _("Failed to open file '%s': %s"),
+                             utf8_filename ? utf8_filename : "???",
+                             g_strerror (errno));
+                g_free (utf8_filename);
+		return NULL;
+        }
+
+	loader = gdk_pixbuf_loader_new ();
+
+	info.width = width;
+	info.height = height;
+        info.preserve_aspect_ratio = preserve_aspect_ratio;
+
+	g_signal_connect (loader, "size-prepared", G_CALLBACK (size_prepared_cb), &info);
+
+	while (!feof (f) && !ferror (f)) {
+		length = fread (buffer, 1, sizeof (buffer), f);
+		if (length > 0)
+			if (!gdk_pixbuf_loader_write (loader, buffer, length, error)) {
+				gdk_pixbuf_loader_close (loader, NULL);
+				fclose (f);
+				g_object_unref (loader);
+				return NULL;
+			}
+	}
+
+	fclose (f);
+
+	if (!gdk_pixbuf_loader_close (loader, error)) {
+		g_object_unref (loader);
+		return NULL;
+	}
+
+	pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
+
+	if (!pixbuf) {
+                gchar *utf8_filename = g_filename_to_utf8 (filename, -1,
+                                                           NULL, NULL, NULL);
+
+		g_object_unref (loader);
+
+                g_set_error (error,
+                             GDK_PIXBUF_ERROR,
+                             GDK_PIXBUF_ERROR_FAILED,
+                             _("Failed to load image '%s': reason not known, probably a corrupt image file"),
+                             utf8_filename ? utf8_filename : "???");
+                g_free (utf8_filename);
+		return NULL;
+	}
+
+	g_object_ref (pixbuf);
+
+	g_object_unref (loader);
+
+	return pixbuf;
+}
