@@ -41,6 +41,7 @@ struct _Icon {
 	DirItem		*item;
 	gchar		*shortcut;	/* Eg: Control + x */
 	MyKey		shortcut_key;	/* Parsed version of shortcut */
+        gchar           *args;          /* Arguments, or NULL if none */
 
 	GtkWidget	*dialog;	/* Current rename box, if any */
 };
@@ -56,5 +57,7 @@ void icon_set_path(Icon *icon, const char *pathname, const char *name);
 gchar *icon_create_uri_list(void);
 void icon_destroy(Icon *icon);
 void icon_set_shortcut(Icon *icon, const gchar *shortcut);
+void icon_set_arguments(Icon *icon, const gchar *args);
+void icon_run(Icon *icon);
 
 #endif /* _ICON_H */
