@@ -1,8 +1,8 @@
 /*
- * $Id: menu.c,v 1.273 2004/01/21 20:03:15 tal197 Exp $
+ * $Id: menu.c,v 1.286 2005/01/16 16:48:01 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -60,6 +60,9 @@
 #include "view_iface.h"
 #include "display.h"
 #include "bookmarks.h"
+#include "panel.h"
+#include "bulk_rename.h"
+#include "xtypes.h"
 
 typedef enum {
 	FILE_COPY_ITEM,
@@ -93,7 +96,7 @@ static GtkWidget *popup_menu = NULL;	/* Currently open menu */
 static gint updating_menu = 0;		/* Non-zero => ignore activations */
 static GList *send_to_paths = NULL;
 
-static Option o_menu_iconsize, o_menu_xterm;
+static Option o_menu_iconsize, o_menu_xterm, o_menu_quick;
 
 /* Static prototypes */
 
@@ -130,6 +133,7 @@ static void reverse_sort(gpointer data, guint action, GtkWidget *widget);
 static void hidden(gpointer data, guint action, GtkWidget *widget);
 static void show_thumbs(gpointer data, guint action, GtkWidget *widget);
 static void refresh(gpointer data, guint action, GtkWidget *widget);
+static void save_settings(gpointer data, guint action, GtkWidget *widget);
 
 static void file_op(gpointer data, FileOp action, GtkWidget *widget);
 
@@ -166,6 +170,7 @@ static GtkWidget	*filer_thumb_menu;	/* The Show Thumbs item */
 static GtkWidget	*filer_new_window;	/* The New Window item */
 static GtkWidget        *filer_new_menu;        /* The New submenu */
 static GtkWidget        *filer_follow_sym;      /* Follow symbolic links item */
+static GtkWidget        *filer_set_type;        /* Set type item */
 
 #undef N_
 #define N_(x) x
@@ -180,8 +185,8 @@ static GtkItemFactoryEntry filer_menu_def[] = {
 {">>" N_("Times"),		NULL, set_with, DETAILS_TIMES, NULL},
 {">" N_("List View"),   	NULL, view_type, VIEW_TYPE_DETAILS, "<StockItem>", ROX_STOCK_SHOW_DETAILS},
 {">",				NULL, NULL, 0, "<Separator>"},
-{">" N_("Bigger Icons"),   	NULL, change_size, 1, "<StockItem>", GTK_STOCK_ZOOM_IN},
-{">" N_("Smaller Icons"),   	NULL, change_size, -1, "<StockItem>", GTK_STOCK_ZOOM_OUT},
+{">" N_("Bigger Icons"),   	"equal", change_size, 1, "<StockItem>", GTK_STOCK_ZOOM_IN},
+{">" N_("Smaller Icons"),   	"minus", change_size, -1, "<StockItem>", GTK_STOCK_ZOOM_OUT},
 {">" N_("Automatic"),   	NULL, change_size_auto, 0, "<ToggleItem>"},
 {">",				NULL, NULL, 0, "<Separator>"},
 {">" N_("Sort by Name"),	NULL, set_sort, SORT_NAME, NULL},
@@ -192,32 +197,35 @@ static GtkItemFactoryEntry filer_menu_def[] = {
 {">" N_("Sort by Group"),	NULL, set_sort, SORT_GROUP, NULL},
 {">" N_("Reversed"),		NULL, reverse_sort, 0, "<ToggleItem>"},
 {">",				NULL, NULL, 0, "<Separator>"},
-{">" N_("Show Hidden"),   	NULL, hidden, 0, "<ToggleItem>"},
+{">" N_("Show Hidden"),   	"<Ctrl>H", hidden, 0, "<ToggleItem>"},
+{">" N_("Filter Files..."),   	NULL, mini_buffer, MINI_FILTER, NULL},
 {">" N_("Show Thumbnails"),	NULL, show_thumbs, 0, "<ToggleItem>"},
 {">" N_("Refresh"),		NULL, refresh, 0, "<StockItem>", GTK_STOCK_REFRESH},
+{">" N_("Save Display Settings..."),	       NULL, save_settings, 0, NULL},
 {N_("File"),			NULL, NULL, 0, "<Branch>"},
-{">" N_("Copy..."),		NULL, file_op, FILE_COPY_ITEM, "<StockItem>", GTK_STOCK_COPY},
+{">" N_("Copy..."),		"<Ctrl>C", file_op, FILE_COPY_ITEM, "<StockItem>", GTK_STOCK_COPY},
 {">" N_("Rename..."),		NULL, file_op, FILE_RENAME_ITEM, NULL},
 {">" N_("Link..."),		NULL, file_op, FILE_LINK_ITEM, NULL},
-{">" N_("Delete"),	    	NULL, file_op, FILE_DELETE, "<StockItem>", GTK_STOCK_DELETE},
+{">" N_("Delete"),	    	"<Ctrl>X", file_op, FILE_DELETE, "<StockItem>", GTK_STOCK_DELETE},
 {">",				NULL, NULL, 0, "<Separator>"},
 {">" N_("Shift Open"),   	NULL, file_op, FILE_OPEN_FILE},
 {">" N_("Open AVFS"),		NULL, file_op, FILE_OPEN_VFS_AVFS, "<StockItem>", GTK_STOCK_OPEN},
 {">" N_("Send To..."),		NULL, file_op, FILE_SEND_TO, NULL},
 {">",				NULL, NULL, 0, "<Separator>"},
-{">" N_("Set Run Action..."),	NULL, file_op, FILE_RUN_ACTION, "<StockItem>", GTK_STOCK_EXECUTE},
+{">" N_("Set Run Action..."),	"asterisk", file_op, FILE_RUN_ACTION, "<StockItem>", GTK_STOCK_EXECUTE},
 {">" N_("Set Icon..."),		NULL, file_op, FILE_SET_ICON, NULL},
-{">" N_("Properties"),		NULL, file_op, FILE_PROPERTIES, "<StockItem>", GTK_STOCK_PROPERTIES},
+{">" N_("Properties"),		"<Ctrl>P", file_op, FILE_PROPERTIES, "<StockItem>", GTK_STOCK_PROPERTIES},
 {">" N_("Count"),		NULL, file_op, FILE_USAGE, NULL},
 {">" N_("Set Type..."),		NULL, file_op, FILE_SET_TYPE, NULL},
 {">" N_("Permissions"),		NULL, file_op, FILE_CHMOD_ITEMS, NULL},
 {">",				NULL, NULL, 0, "<Separator>"},
-{">" N_("Find"),		NULL, file_op, FILE_FIND, "<StockItem>", GTK_STOCK_FIND},
+{">" N_("Find"),		"<Ctrl>F", file_op, FILE_FIND, "<StockItem>", GTK_STOCK_FIND},
 {N_("Select"),	    		NULL, NULL, 0, "<Branch>"},
-{">" N_("Select All"),	    	NULL, select_all, 0, NULL},
+{">" N_("Select All"),	    	"<Ctrl>A", select_all, 0, NULL},
 {">" N_("Clear Selection"),	NULL, clear_selection, 0, NULL},
 {">" N_("Invert Selection"),	NULL, invert_selection, 0, NULL},
-{">" N_("Select If..."),	NULL, mini_buffer, MINI_SELECT_IF, NULL},
+{">" N_("Select by Name..."),	"period", mini_buffer, MINI_SELECT_BY_NAME, NULL},
+{">" N_("Select If..."),	"<Shift>question", mini_buffer, MINI_SELECT_IF, NULL},
 {N_("Options..."),		NULL, menu_show_options, 0, "<StockItem>", GTK_STOCK_PREFERENCES},
 {N_("New"),			NULL, NULL, 0, "<Branch>"},
 {">" N_("Directory"),		NULL, new_directory, 0, NULL},
@@ -226,17 +234,17 @@ static GtkItemFactoryEntry filer_menu_def[] = {
 {">" N_("Parent, New Window"), 	NULL, open_parent, 0, "<StockItem>", GTK_STOCK_GO_UP},
 {">" N_("Parent, Same Window"), NULL, open_parent_same, 0, NULL},
 {">" N_("New Window"),		NULL, new_window, 0, NULL},
-{">" N_("Home Directory"),	NULL, home_directory, 0, "<StockItem>", GTK_STOCK_HOME},
+{">" N_("Home Directory"),	"<Ctrl>Home", home_directory, 0, "<StockItem>", GTK_STOCK_HOME},
 {">" N_("Show Bookmarks"),	"<Ctrl>B", show_bookmarks, 0, "<StockItem>", ROX_STOCK_BOOKMARKS},
 {">" N_("Follow Symbolic Links"),	NULL, follow_symlinks, 0, NULL},
-{">" N_("Resize Window"),	NULL, resize, 0, NULL},
+{">" N_("Resize Window"),	"<Ctrl>E", resize, 0, NULL},
 /* {">" N_("New, As User..."),	NULL, new_user, 0, NULL}, */
 
-{">" N_("Close Window"),	NULL, close_window, 0, "<StockItem>", GTK_STOCK_CLOSE},
+{">" N_("Close Window"),	"<Ctrl>Q", close_window, 0, "<StockItem>", GTK_STOCK_CLOSE},
 {">",				NULL, NULL, 0, "<Separator>"},
 {">" N_("Enter Path..."),	"slash", mini_buffer, MINI_PATH, NULL},
-{">" N_("Shell Command..."),	NULL, mini_buffer, MINI_SHELL, NULL},
-{">" N_("Xterm Here"),		NULL, xterm_here, FALSE, NULL},
+{">" N_("Shell Command..."),	"<Shift>exclam", mini_buffer, MINI_SHELL, NULL},
+{">" N_("Xterm Here"),		"grave", xterm_here, FALSE, NULL},
 {">" N_("Switch to xterm"),	NULL, xterm_here, TRUE, NULL},
 {N_("Help"),			NULL, NULL, 0, "<Branch>"},
 {">" N_("About ROX-Filer..."),	NULL, menu_rox_help, HELP_ABOUT, NULL},
@@ -285,6 +293,8 @@ gboolean ensure_filer_menu(void)
 	GET_SSMENU_ITEM(filer_auto_size_menu, "filer", "Display", "Automatic");
 	GET_SSMENU_ITEM(filer_thumb_menu, "filer", "Display",
 							"Show Thumbnails");
+	GET_SSMENU_ITEM(item, "filer", "File", "Set Type...");
+	filer_set_type = GTK_BIN(item)->child;
 
 	GET_SMENU_ITEM(filer_new_menu, "filer", "New");
 	GET_SSMENU_ITEM(item, "filer", "Window", "Follow Symbolic Links");
@@ -327,6 +337,7 @@ void menu_init(void)
 
 	option_add_string(&o_menu_xterm, "menu_xterm", "xterm");
 	option_add_int(&o_menu_iconsize, "menu_iconsize", MIS_SMALL);
+	option_add_int(&o_menu_quick, "menu_quick", FALSE);
 	option_add_saver(save_menus);
 
 	option_register_widget("menu-set-keys", set_keys_button);
@@ -372,7 +383,8 @@ static void menuitem_no_shortcuts(GtkWidget *item)
 /* Shade items that only work on single files */
 static void shade_file_menu_items(gboolean shaded)
 {
-	menu_set_items_shaded(filer_file_menu, shaded, 0, 3);
+	menu_set_items_shaded(filer_file_menu, shaded, 0, 1);
+	menu_set_items_shaded(filer_file_menu, shaded, 2, 1);
 	menu_set_items_shaded(filer_file_menu, shaded, 5, 2);
 	menu_set_items_shaded(filer_file_menu, shaded, 9, 2);
 }
@@ -422,19 +434,22 @@ GtkWidget *make_send_to_item(DirItem *ditem, const char *label,
 {
 	GtkWidget *item;
 	
-	if (ditem->image && style != MIS_NONE)
+	if (style != MIS_NONE && di_image(ditem))
 	{
 		GdkPixbuf *pixbuf;
+		MaskedPixmap *image;
+
+		image = di_image(ditem);
 
 		switch (style)
 		{
 			case MIS_LARGE:
-				pixbuf = ditem->image->pixbuf;
+				pixbuf = image->pixbuf;
 				break;
 			default:
-				if (!ditem->image->sm_pixbuf)
-					pixmap_make_small(ditem->image);
-				pixbuf = ditem->image->sm_pixbuf;
+				if (!image->sm_pixbuf)
+					pixmap_make_small(image);
+				pixbuf = image->sm_pixbuf;
 				break;
 		}
 
@@ -739,7 +754,7 @@ void show_filer_menu(FilerWindow *filer_window, GdkEvent *event, ViewIter *iter)
 				break;
 			case 1:
 				item = filer_selected_item(filer_window);
-				if (!item->image)
+				if (item->base_type == TYPE_UNKNOWN)
 					dir_update_item(filer_window->directory,
 							item->leafname);
 				shade_file_menu_items(FALSE);
@@ -777,15 +792,22 @@ void show_filer_menu(FilerWindow *filer_window, GdkEvent *event, ViewIter *iter)
 			!o_unique_filer_windows.int_value);
 	gtk_widget_set_sensitive(filer_follow_sym,
 		strcmp(filer_window->sym_path, filer_window->real_path) != 0);
+	gtk_widget_set_sensitive(filer_set_type,
+				 xtype_supported(filer_window->real_path));
 
-	popup_menu = (state & GDK_CONTROL_MASK)
-				? filer_file_menu
-				: filer_menu;
+	if (n_selected && o_menu_quick.int_value) 
+		popup_menu = (state & GDK_CONTROL_MASK)
+					? filer_menu
+					: filer_file_menu;
+	else 
+		popup_menu = (state & GDK_CONTROL_MASK)
+					? filer_file_menu
+					: filer_menu;
 
 	updating_menu--;
 	
 	show_popup_menu(popup_menu, event,
-			popup_menu == filer_file_menu ? (n_added + 7) : 1);
+			popup_menu == filer_file_menu ? n_added : 1);
 }
 
 static void menu_closed(GtkWidget *widget)
@@ -939,7 +961,8 @@ static void hidden(gpointer data, guint action, GtkWidget *widget)
 
 	g_return_if_fail(window_with_focus != NULL);
 
-	display_set_hidden(window_with_focus, !window_with_focus->show_hidden);
+	display_set_hidden(window_with_focus,
+			   !window_with_focus->show_hidden);
 }
 
 static void show_thumbs(gpointer data, guint action, GtkWidget *widget)
@@ -959,6 +982,13 @@ static void refresh(gpointer data, guint action, GtkWidget *widget)
 	filer_refresh(window_with_focus);
 }
 
+static void save_settings(gpointer data, guint action, GtkWidget *widget)
+{
+	g_return_if_fail(window_with_focus != NULL);
+
+	filer_save_settings(window_with_focus);
+}
+
 static void delete(FilerWindow *filer_window)
 {
 	GList *paths;
@@ -985,9 +1015,26 @@ static void chmod_items(FilerWindow *filer_window)
 
 static void set_type_items(FilerWindow *filer_window)
 {
-	GList *paths;
+	GList *paths, *p;
+	int npass=0, nfail=0;
+	
 	paths = filer_selected_items(filer_window);
-	action_settype(paths, FALSE, NULL);
+	for(p=paths; p; p=g_list_next(p)) {
+		if(xtype_supported((const char *) p->data))
+			npass++;
+		else
+			nfail++;
+	}
+	if(npass==0) 
+		report_error(_("Extended attributes, used to store types, are not supported for this "
+				"file or files.\n"
+				"This may be due to lack of support from the filesystem or the C library, "
+				"or it may simply be that the filesystem needs to be mounted with "
+				"the right mount option ('user_xattr' on Linux)."));
+	else if(nfail>0)
+		report_error(_("Setting type not supported for some of these files"));
+	if(npass>0)
+		action_settype(paths, FALSE, NULL);
 	destroy_glist(&paths);
 }
 
@@ -1517,7 +1564,45 @@ static void show_send_to_menu(GList *paths, GdkEvent *event)
 		diritem_free(item);
 	}
 	else
+	{
+		GList *rover;
+		gboolean same=TRUE, same_media=TRUE;
+		MIME_type *type=NULL;
+		DirItem	*item;
+		
+		item = diritem_new("");
+		for(rover=paths; rover; rover=g_list_next(rover))
+		{
+			diritem_restat(rover->data, item, NULL);
+			if(!type)
+				type=item->mime_type;
+			else
+			{
+				if(type!=item->mime_type)
+				{
+					same=FALSE;
+					if(strcmp(type->media_type,
+						  item->mime_type->media_type)!=0)
+					{
+						same_media=FALSE;
+						break;
+					}
+				}
+			}
+		}
+		diritem_free(item);
+
+		if(type)
+		{
+			if(same)
+				add_sendto(menu, type->media_type,
+					   type->subtype);
+			if(same_media)
+				add_sendto(menu, type->media_type, NULL);
+		}
+		
 		add_sendto(menu, "group", NULL);
+	}
 	
 	add_sendto(menu, NULL, NULL);
 
@@ -1817,6 +1902,22 @@ static void file_op(gpointer data, FileOp action, GtkWidget *unused)
 			destroy_glist(&items);
 			return;
 		}
+		case FILE_RENAME_ITEM:
+			if (n_selected > 1)
+			{
+				GList *items = NULL;
+				ViewIter iter;
+
+				view_get_iter(window_with_focus->view, &iter, VIEW_ITER_SELECTED);
+				while ((item = iter.next(&iter)))
+					items = g_list_prepend(items, item->leafname);
+				items = g_list_reverse(items);
+
+				bulk_rename(window_with_focus->sym_path, items);
+				g_list_free(items);
+				return;
+			}
+			break;	/* Not a bulk rename... see below */
 		default:
 			break;
 	}
@@ -1836,7 +1937,7 @@ static void file_op(gpointer data, FileOp action, GtkWidget *unused)
 	g_return_if_fail(item != NULL);
 	/* iter may be passed to filer_openitem... */
 
-	if (!item->image)
+	if (item->base_type == TYPE_UNKNOWN)
 		item = dir_update_item(window_with_focus->directory,
 					item->leafname);
 
@@ -1851,17 +1952,17 @@ static void file_op(gpointer data, FileOp action, GtkWidget *unused)
 	switch (action)
 	{
 		case FILE_COPY_ITEM:
-			src_dest_action_item(path, item->image,
+			src_dest_action_item(path, di_image(item),
 					_("Copy"), copy_cb,
 					GDK_ACTION_COPY);
 			break;
 		case FILE_RENAME_ITEM:
-			src_dest_action_item(path, item->image,
+			src_dest_action_item(path, di_image(item),
 					_("Rename"), rename_cb,
 					GDK_ACTION_MOVE);
 			break;
 		case FILE_LINK_ITEM:
-			src_dest_action_item(path, item->image,
+			src_dest_action_item(path, di_image(item),
 					_("Symlink"), link_cb,
 					GDK_ACTION_LINK);
 			break;
