@@ -1,8 +1,8 @@
 /*
- * $Id: panel.c,v 1.113 2004/04/30 16:36:41 tal197 Exp $
+ * $Id: panel.c,v 1.132 2005/01/16 16:48:01 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -98,8 +98,8 @@ static int panel_delete(GtkWidget *widget, GdkEvent *event, Panel *panel);
 static void panel_destroyed(GtkWidget *widget, Panel *panel);
 static const char *pan_from_file(gchar *line);
 static gint icon_button_release(GtkWidget *widget,
-			        GdkEventButton *event,
-			        PanelIcon *pi);
+				GdkEventButton *event,
+				PanelIcon *pi);
 static gint icon_button_press(GtkWidget *widget,
 			      GdkEventButton *event,
 			      PanelIcon *pi);
@@ -121,42 +121,43 @@ static void panel_post_resize(GtkWidget *box,
 			GtkRequisition *req, Panel *panel);
 static void drag_set_panel_dest(PanelIcon *pi);
 static void add_uri_list(GtkWidget          *widget,
-                         GdkDragContext     *context,
-                         gint               x,
-                         gint               y,
-                         GtkSelectionData   *selection_data,
-                         guint              info,
-                         guint32            time,
+			 GdkDragContext     *context,
+			 gint               x,
+			 gint               y,
+			 GtkSelectionData   *selection_data,
+			 guint              info,
+			 guint32            time,
 			 Panel		    *panel);
 static void panel_add_item(Panel *panel,
 			   const gchar *path,
 			   const gchar *name,
 			   gboolean after,
-			   const gchar *shortcut);
+			   const gchar *shortcut,
+			   const gchar *args);
 static gboolean panel_drag_motion(GtkWidget	*widget,
-                            GdkDragContext	*context,
-                            gint		x,
-                            gint		y,
-                            guint		time,
+			    GdkDragContext	*context,
+			    gint		x,
+			    gint		y,
+			    guint		time,
 			    Panel		*panel);
 static gboolean insert_drag_motion(GtkWidget	*widget,
-                            GdkDragContext	*context,
-                            gint		x,
-                            gint		y,
-                            guint		time,
+			    GdkDragContext	*context,
+			    gint		x,
+			    gint		y,
+			    guint		time,
 			    Panel		*panel);
 static gboolean drag_motion(GtkWidget		*widget,
-                            GdkDragContext	*context,
-                            gint		x,
-                            gint		y,
-                            guint		time,
+			    GdkDragContext	*context,
+			    gint		x,
+			    gint		y,
+			    guint		time,
 			    PanelIcon		*pi);
 static void panel_drag_leave(GtkWidget	*widget,
-                       GdkDragContext	*context,
+		       GdkDragContext	*context,
 		       guint32		time,
 		       Panel		*panel);
 static void drag_leave(GtkWidget	*widget,
-                       GdkDragContext	*context,
+		       GdkDragContext	*context,
 		       guint32		time,
 		       Icon	*icon);
 static GtkWidget *make_insert_frame(Panel *panel);
@@ -197,6 +198,7 @@ static void panel_update(Panel *panel);
 static gboolean panel_check_xinerama(void);
 static GList *build_monitor_number(Option *option,
 					xmlNode *node, guchar *label);
+static gboolean may_autoscroll(Panel *panel);
 
 
 static GtkWidget *dnd_highlight = NULL; /* (stops flickering) */
@@ -208,6 +210,8 @@ static Option o_panel_style;
 static Option o_panel_width;
 static Option o_panel_xinerama;
 static Option o_panel_monitor;
+static Option o_panel_avoid;
+static Option o_panel_is_dock;
 
 static gint panel_monitor = -1;
 GdkRectangle panel_geometry;
@@ -226,6 +230,9 @@ void panel_init(void)
 	option_add_int(&o_panel_xinerama, "panel_xinerama", 0);
 	option_add_int(&o_panel_monitor, "panel_monitor", 0);
 
+	option_add_int(&o_panel_avoid, "panel_avoid", TRUE);
+	option_add_int(&o_panel_is_dock, "panel_is_dock", FALSE);
+
 	option_add_notify(panel_style_changed);
 
 	option_register_widget("monitor-number", build_monitor_number);
@@ -352,11 +359,14 @@ Panel *panel_new(const gchar *name, PanelSide side)
 	g_object_set_data(G_OBJECT(frame), "after", "yes");
 	gtk_box_pack_start(GTK_BOX(box), frame, TRUE, TRUE, 4);
 
+	if (o_panel_is_dock.int_value)
+		gtk_window_set_type_hint(GTK_WINDOW(panel->window),
+				GDK_WINDOW_TYPE_HINT_DOCK);	
+
 	gtk_widget_realize(panel->window);
 	make_panel_window(panel->window);
-	if (!(gtk_major_version == 2 && gtk_minor_version == 0))
-		gtk_window_stick(GTK_WINDOW(panel->window));
-
+	gtk_window_stick(GTK_WINDOW(panel->window));
+	
 	gtk_widget_show_all(align);
 	
 	loading_panel = panel;
@@ -382,12 +392,12 @@ Panel *panel_new(const gchar *name, PanelSide side)
 		/* Don't scare users with an empty panel... */
 		guchar *apps;
 		
-		panel_add_item(panel, "~", "Home", FALSE, NULL);
+		panel_add_item(panel, "~", "Home", FALSE, NULL, NULL);
 
 		apps = pathdup(make_path(app_dir, ".."));
 		if (apps)
 		{
-			panel_add_item(panel, apps, "Apps", FALSE, NULL);
+			panel_add_item(panel, apps, "Apps", FALSE, NULL, NULL);
 			g_free(apps);
 		}
 	}
@@ -402,31 +412,11 @@ Panel *panel_new(const gchar *name, PanelSide side)
 	g_signal_connect(panel->window, "size-allocate",
 			G_CALLBACK(reposition_panel), panel);
 
-	/* Stop windows from maximising over us completely, so that the
-	 * auto-raise stuff works...
-	 */
-	{
-		guint32	wm_strut[] = {0, 0, 0, 0};
-
-		if (panel->side == PANEL_LEFT)
-			wm_strut[0] = 2;
-		else if (panel->side == PANEL_RIGHT)
-			wm_strut[1] = 2;
-		else if (panel->side == PANEL_TOP)
-			wm_strut[2] = 2;
-		else
-			wm_strut[3] = 2;
-
-		gdk_property_change(panel->window->window,
-				gdk_atom_intern("_NET_WM_STRUT", FALSE),
-				gdk_atom_intern("CARDINAL", FALSE),
-				32, GDK_PROP_MODE_REPLACE,
-				(gchar *) &wm_strut, 4);
-	}
-
 	number_of_windows++;
 	gdk_window_lower(panel->window->window);
 	gtk_widget_show(panel->window);
+	/* This has no effect until after window is showing; GTK+ bug? */
+	keep_below(panel->window->window, TRUE);
 
 	{
 		GdkWindow *pinboard;
@@ -447,7 +437,7 @@ gboolean panel_add(PanelSide side,
 	
 	g_return_val_if_fail(current_panel[side] != NULL, FALSE);
 
-	panel_add_item(current_panel[side], path, label, after, NULL);
+	panel_add_item(current_panel[side], path, label, after, NULL, NULL);
 
 	return TRUE;
 }
@@ -479,10 +469,17 @@ void panel_update_size(void)
 {
 	int i;
 
+	panel_check_xinerama();
+
 	for (i = 0; i < PANEL_NUMBER_OF_SIDES; i++)
 	{
 		if (current_panel[i])
-			panel_update(current_panel[i]);
+		{
+			reposition_panel(current_panel[i]->window,
+					&current_panel[i]->window->allocation,
+					current_panel[i]);
+			gtk_widget_queue_resize(current_panel[i]->window);
+		}
 	}
 }
 
@@ -526,7 +523,7 @@ static void panel_destroyed(GtkWidget *widget, Panel *panel)
 static void panel_load_side(Panel *panel, xmlNodePtr side, gboolean after)
 {
 	xmlNodePtr node;
-	char	   *label, *path, *shortcut;
+	char	   *label, *path, *shortcut, *args;
 
 	for (node = side->xmlChildrenNode; node; node = node->next)
 	{
@@ -542,12 +539,14 @@ static void panel_load_side(Panel *panel, xmlNodePtr side, gboolean after)
 		if (!path)
 			path = g_strdup("<missing path>");
 		shortcut = xmlGetProp(node, "shortcut");
+		args = xmlGetProp(node, "args");
 
-		panel_add_item(panel, path, label, after, shortcut);
+		panel_add_item(panel, path, label, after, shortcut, args);
 
 		g_free(path);
 		g_free(label);
 		g_free(shortcut);
+		g_free(args);
 	}
 }
 
@@ -581,7 +580,8 @@ static const char *pan_from_file(gchar *line)
 	else
 		leaf = NULL;
 	
-	panel_add_item(loading_panel, sep + 1, leaf, sep[0] == '>', NULL);
+	panel_add_item(loading_panel, sep + 1, leaf, sep[0] == '>',
+		       NULL, NULL);
 
 	g_free(leaf);
 
@@ -592,7 +592,8 @@ static gboolean icon_pointer_in(GtkWidget *widget,
 				GdkEventCrossing *event,
 				Icon *icon)
 {
-	gtk_widget_set_state(widget, GTK_STATE_PRELIGHT);
+	gtk_widget_set_state(widget,
+		icon->selected ? GTK_STATE_SELECTED : GTK_STATE_PRELIGHT);
 
 	return 0;
 }
@@ -601,7 +602,8 @@ static gboolean icon_pointer_out(GtkWidget *widget,
 				 GdkEventCrossing *event,
 				 Icon *icon)
 {
-	gtk_widget_set_state(widget, GTK_STATE_NORMAL);
+	gtk_widget_set_state(widget,
+		icon->selected ? GTK_STATE_SELECTED : GTK_STATE_NORMAL);
 
 	return 0;
 }
@@ -650,8 +652,9 @@ static void panel_icon_set_tip(PanelIcon *pi)
 	}
 	else if ((!panel_want_show_text(pi)) && !pi->socket)
 	{
-		gtk_tooltips_set_tip(tooltips, pi->widget,
-				icon->item->leafname, NULL);
+		if (icon->item->leafname && icon->item->leafname[0])
+			gtk_tooltips_set_tip(tooltips, pi->widget,
+					icon->item->leafname, NULL);
 	}
 	else
 		gtk_tooltips_set_tip(tooltips, pi->widget, NULL, NULL);
@@ -669,7 +672,8 @@ static void panel_add_item(Panel *panel,
 			   const gchar *path,
 			   const gchar *name,
 			   gboolean after,
-			   const gchar *shortcut)
+			   const gchar *shortcut,
+			   const gchar *args)
 {
 	GtkWidget	*widget;
 	PanelIcon	*pi;
@@ -741,14 +745,76 @@ static void panel_add_item(Panel *panel,
 	}
 
 	icon_set_shortcut(icon, shortcut);
+	icon_set_arguments(icon, args);
 
 	if (!loading_panel)
 		panel_save(panel);
-		
+
 	panel_icon_set_tip(pi);
 	gtk_widget_show(widget);
 }
 
+static gboolean remove_item_from_side(GtkWidget *container, const gchar *path,
+				      const gchar *label)
+{
+	GList *kids, *next;
+	gboolean found = FALSE;
+
+	kids = gtk_container_get_children(GTK_CONTAINER(container));
+
+	for (next = kids; next; next = next->next)
+	{
+		Icon    *icon;
+		icon = g_object_get_data(G_OBJECT(next->data), "icon");
+		if (!icon)
+			continue;
+
+		if (strcmp(path, icon->src_path) == 0 &&
+		    (!label || strcmp(label, icon->item->leafname)==0))
+		{
+			icon_destroy(icon);
+			found = TRUE;
+			break;
+		}
+	}
+	
+	g_list_free(kids);
+
+	return found;
+}
+
+/* Remove an item with this path. If more than one item matches, only
+ * one is removed. If label is not NULL then it must also match the item.
+ * Returns TRUE if an item was successfully removed.
+ */
+gboolean panel_remove_item(PanelSide side, const gchar *path,
+			   const gchar *label)
+{
+	Panel *panel;
+
+	g_return_val_if_fail(side >= 0 && side < PANEL_NUMBER_OF_SIDES, FALSE);
+	g_return_val_if_fail(path != NULL, FALSE);
+
+	panel = current_panel[side];
+	if (!panel)
+	{
+		g_warning("No panel on this side of the screen!");
+		return FALSE;
+	}
+
+	if (remove_item_from_side(panel->before, path, label) ||
+	    remove_item_from_side(panel->after, path, label))
+	{
+		panel_save(panel);
+		panel_update(panel);
+		return TRUE;
+	}
+
+
+	g_warning("Panel item '%s' not found", path);
+	return FALSE;
+}
+
 /* Called when Gtk+ wants to know how much space an icon needs.
  * 'req' is already big enough for the label, if shown.
  */
@@ -770,7 +836,7 @@ static void size_request(GtkWidget *widget, GtkRequisition *req, PanelIcon *pi)
 	if (pi->image)
 		g_object_unref(pi->image);
 
-	pi->image = scale_pixbuf(icon->item->image->src_pixbuf,
+	pi->image = scale_pixbuf(di_image(icon->item)->src_pixbuf,
 			MAX(20, max_width), MAX(20, max_height));
 
 	image_width = gdk_pixbuf_get_width(pi->image);
@@ -815,7 +881,7 @@ static gint draw_icon(GtkWidget *widget, GdkRectangle *badarea, PanelIcon *pi)
 	int		text_height = 0;
 
 	gdk_drawable_get_size(widget->window, &area.width, &area.height);
-	
+
 	if (panel_want_show_text(pi))
 		text_height = pi->label->requisition.height;
 	
@@ -897,7 +963,7 @@ static void perform_action(Panel *panel, PanelIcon *pi, GdkEventButton *event)
 		case ACT_OPEN_ITEM:
 			dnd_motion_ungrab();
 			wink_widget(pi->widget);
-			run_diritem(icon->path, icon->item, NULL, NULL, FALSE);
+			icon_run(icon);
 			break;
 		case ACT_EDIT_ITEM:
 			dnd_motion_ungrab();
@@ -964,8 +1030,8 @@ static gint panel_button_press(GtkWidget *widget,
 }
 
 static gint icon_button_release(GtkWidget *widget,
-			        GdkEventButton *event,
-			        PanelIcon *pi)
+				GdkEventButton *event,
+				PanelIcon *pi)
 {
 	if (pi->socket && event->button == 1)
 		return FALSE;	/* Restart button */
@@ -996,6 +1062,7 @@ static void reposition_panel(GtkWidget *window,
 {
 	int		x = panel_geometry.x;
 	int		y = panel_geometry.y;
+	int		thickness;
 	PanelSide	side = panel->side;
 
 	if (side == PANEL_LEFT || side == PANEL_RIGHT)
@@ -1025,6 +1092,135 @@ static void reposition_panel(GtkWidget *window,
 			gtk_widget_queue_resize(
 					current_panel[PANEL_LEFT]->window);
 	}
+
+	/* Stop windows from maximising over all/part of us */
+	{
+		struct {
+			guint32 left, right, top, bottom;
+			guint32 left_start_y, left_end_y;
+			guint32 right_start_y, right_end_y;
+			guint32 top_start_x, top_end_x;
+			guint32 bottom_start_x, bottom_end_x;
+		} strut = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+
+		if (o_panel_avoid.int_value == FALSE)
+			thickness = 2;
+		else if (panel->side == PANEL_TOP ||
+			 panel->side == PANEL_BOTTOM)
+			thickness = alloc->height;
+		else
+			thickness = alloc->width;
+
+		switch (panel->side)
+		{
+			case PANEL_LEFT:
+				if (!o_panel_xinerama.int_value ||
+					!monitor_adjacent
+					[o_panel_monitor.int_value].left)
+				{
+					strut.left = panel_geometry.x +
+						thickness;
+					strut.left_start_y = panel_geometry.y;
+					strut.left_end_y = panel_geometry.y +
+						panel_geometry.height - 1;
+				}
+				/* else there is (part of) a monitor
+				 * to the left */
+				else
+				{
+					thickness = 0;
+				}
+				break;
+			case PANEL_RIGHT:
+				if (!o_panel_xinerama.int_value ||
+					!monitor_adjacent
+					[o_panel_monitor.int_value].right)
+				{
+					/* RHS of monitor might not abut edge
+					 * of total virtual screen */
+					strut.right = screen_width -
+						panel_geometry.x -
+						panel_geometry.width +
+						thickness;
+					strut.right_start_y = panel_geometry.y;
+					strut.right_end_y = panel_geometry.y +
+						panel_geometry.height - 1;
+				}
+				/* else there is (part of) a monitor
+				 * to the right */
+				else
+				{
+					thickness = 0;
+				}
+				break;
+			case PANEL_TOP:
+				if (!o_panel_xinerama.int_value ||
+					!monitor_adjacent
+					[o_panel_monitor.int_value].top)
+				{
+					strut.top = panel_geometry.y +
+						thickness;
+					strut.top_start_x = panel_geometry.x;
+					strut.top_end_x = panel_geometry.x +
+						panel_geometry.width - 1;
+				}
+				/* else there is (part of) a monitor above */
+				else
+				{
+					thickness = 0;
+				}
+				break;
+			default:	/* PANEL_BOTTOM */
+				if (!o_panel_xinerama.int_value ||
+					!monitor_adjacent
+					[o_panel_monitor.int_value].bottom)
+				{
+					/* Bottom of monitor might not abut
+					 * edge of total virtual screen */
+					strut.bottom = screen_height -
+						panel_geometry.y -
+						panel_geometry.height +
+						thickness;
+					strut.bottom_start_x = panel_geometry.x;
+					strut.bottom_end_x = panel_geometry.x +
+						panel_geometry.width - 1;
+				}
+				/* else there is (part of) a monitor below */
+				else
+				{
+					thickness = 0;
+				}
+				break;
+		}
+
+		if (thickness)
+		{
+			/* Set full-width strut as well as partial in case
+			 * partial isn't supported by wm */
+			gdk_property_change(panel->window->window,
+					gdk_atom_intern("_NET_WM_STRUT",
+						FALSE),
+					gdk_atom_intern("CARDINAL", FALSE),
+					32, GDK_PROP_MODE_REPLACE,
+					(gchar *) &strut, 4);
+			gdk_property_change(panel->window->window,
+					gdk_atom_intern("_NET_WM_STRUT_PARTIAL",
+						FALSE),
+					gdk_atom_intern("CARDINAL", FALSE),
+					32, GDK_PROP_MODE_REPLACE,
+					(gchar *) &strut, 12);
+		}
+		else
+		{
+			gdk_property_delete(panel->window->window,
+					gdk_atom_intern("_NET_WM_STRUT_PARTIAL",
+						FALSE));
+			gdk_property_delete(panel->window->window,
+					gdk_atom_intern("_NET_WM_STRUT",
+						FALSE));
+		}
+	}
+
 }
 
 /* Same as drag_set_dest(), but for panel icons */
@@ -1040,10 +1236,10 @@ static void drag_set_panel_dest(PanelIcon *pi)
 }
 
 static gboolean drag_motion(GtkWidget		*widget,
-                            GdkDragContext	*context,
-                            gint		x,
-                            gint		y,
-                            guint		time,
+			    GdkDragContext	*context,
+			    gint		x,
+			    gint		y,
+			    guint		time,
 			    PanelIcon		*pi)
 {
 	GdkDragAction	action = context->suggested_action;
@@ -1056,6 +1252,11 @@ static gboolean drag_motion(GtkWidget		*widget,
 				&panel_x, &panel_y, NULL);
 	motion_may_raise(pi->panel, panel_x, panel_y);
 
+	/* Should we scroll the panel when dragging? */
+	if (motion_state != MOTION_REPOSITION)
+		if (pi->panel->autoscroll_speed == 0)
+			may_autoscroll(pi->panel);
+
 	if (icon->selected)
 		goto out;	/* Can't drag a selection to itself */
 
@@ -1113,12 +1314,12 @@ out:
 
 
 static void add_uri_list(GtkWidget          *widget,
-                         GdkDragContext     *context,
-                         gint               x,
-                         gint               y,
-                         GtkSelectionData   *selection_data,
-                         guint              info,
-                         guint32            time,
+			 GdkDragContext     *context,
+			 gint               x,
+			 gint               y,
+			 GtkSelectionData   *selection_data,
+			 guint              info,
+			 guint32            time,
 			 Panel		    *panel)
 {
 	gboolean after = FALSE;
@@ -1138,10 +1339,10 @@ static void add_uri_list(GtkWidget          *widget,
 	{
 		guchar	*path;
 
-		path = get_local_path((guchar *) next->data);
+		path = get_local_path((EscapedPath *) next->data);
 
 		if (path) {
-			panel_add_item(panel, path, NULL, after, NULL);
+			panel_add_item(panel, path, NULL, after, NULL, NULL);
 			g_free(path);
 		}
 	}
@@ -1161,7 +1362,7 @@ static void drag_end(GtkWidget *widget,
 }
 
 static void drag_leave(GtkWidget	*widget,
-                       GdkDragContext	*context,
+		       GdkDragContext	*context,
 		       guint32		time,
 		       Icon	*icon)
 {
@@ -1201,6 +1402,8 @@ static void make_widgets(xmlNodePtr side, GList *widgets)
 		xmlSetProp(tree, "label", icon->item->leafname);
 		if (icon->shortcut)
 			xmlSetProp(tree, "shortcut", icon->shortcut);
+		if (icon->args)
+			xmlSetProp(tree, "args", icon->args);
 	}
 	
 	if (widgets)
@@ -1299,6 +1502,10 @@ static gint panel_leave_event(GtkWidget *widget,
 	if (event->mode != GDK_CROSSING_NORMAL)
 		return FALSE;	/* Grab for menu, DnD, etc */
 
+	keep_below(panel->window->window, TRUE);
+
+	/* Shouldn't need this as well as keep_below but some WMs don't
+	 * automatically lower as soon as the hint is set */ 
 	pinboard = pinboard_get_window();
 	window_put_just_above(panel->window->window, pinboard);
 
@@ -1320,7 +1527,13 @@ static void motion_may_raise(Panel *panel, int x, int y)
 		raise = x == panel->window->allocation.width - 1;
 
 	if (raise)
+	{
+		keep_below(panel->window->window, FALSE);
+
+		/* Shouldn't need this as well as keep_below but some WMs don't
+		 * automatically raise as soon as the hint is set */ 
 		gdk_window_raise(panel->window->window);
+	}
 }
 
 static gboolean may_autoscroll(Panel *panel)
@@ -1443,7 +1656,7 @@ static gint icon_motion_event(GtkWidget *widget,
 		x += w;
 		y += h;
 
-		if (val >= (horz ? x : y))
+		if (val >= (horz ? x : y)-1)
 		{
 			if (next == panel->gap)
 				dir = +2;
@@ -1631,6 +1844,11 @@ static void run_applet(PanelIcon *pi)
 				8, GDK_PROP_MODE_REPLACE,
 				pos, strlen(pos));
 		g_free(pos);
+
+		/* Ensure that the properties are set before starting the
+		 * applet.
+		 */
+		gdk_flush();
 	}
 
 	g_object_set_data(G_OBJECT(pi->widget), "icon", pi);
@@ -1789,9 +2007,10 @@ static void panel_style_changed(void)
 		}
 	}
 
-	if (o_panel_xinerama.has_changed || o_panel_monitor.has_changed)
+	if (o_panel_xinerama.has_changed || o_panel_monitor.has_changed ||
+	    o_panel_avoid.has_changed)
 	{
-		if (panel_check_xinerama())
+		if (panel_check_xinerama() || o_panel_avoid.has_changed)
 		{
 			for (i = 0; i < PANEL_NUMBER_OF_SIDES; i++)
 			{
@@ -1877,6 +2096,9 @@ static void panel_remove_items(void)
 
 static void panel_icon_redraw(Icon *icon)
 {
+	gtk_widget_set_state(((PanelIcon *) icon)->widget,
+			icon->selected ? GTK_STATE_SELECTED
+				       : GTK_STATE_NORMAL);
 	gtk_widget_queue_draw(PANEL_ICON(icon)->widget);
 }
 
@@ -1997,37 +2219,47 @@ static void panel_position_menu(GtkMenu *menu, gint *x, gint *y,
 	int		*pos = (int *) data;
 	GtkRequisition 	requisition;
 	int		margin = pos[2];
+	int		mon = pos[3];
+	int		mon_right = monitor_geom[mon].x +
+				monitor_geom[mon].width;
+	int		mon_bottom = monitor_geom[mon].y +
+				monitor_geom[mon].height;
 
 	gtk_widget_size_request(GTK_WIDGET(menu), &requisition);
 
 	if (pos[0] == -1)
-		*x = screen_width - margin - requisition.width;
+		*x = mon_right - margin - requisition.width;
 	else if (pos[0] == -2)
-		*x = margin;
+		*x = monitor_geom[mon].x + margin;
 	else
 		*x = pos[0] - (requisition.width >> 2);
 		
 	if (pos[1] == -1)
-		*y = screen_height - margin - requisition.height;
+		*y = mon_bottom - margin - requisition.height;
 	else if (pos[1] == -2)
-		*y = margin;
+		*y = monitor_geom[mon].y + margin;
 	else
 		*y = pos[1] - (requisition.height >> 2);
 
-	*x = CLAMP(*x, 0, screen_width - requisition.width);
-	*y = CLAMP(*y, 0, screen_height - requisition.height);
+	*x = CLAMP(*x, 0, mon_right - requisition.width);
+	*y = CLAMP(*y, 0, mon_bottom - requisition.height);
 
 	*push_in = FALSE;
 }
 
 static void panel_show_menu(GdkEventButton *event, PanelIcon *pi, Panel *panel)
 {
-	PanelSide side = panel->side;
-	int pos[3];
+	PanelSide	side = panel->side;
+	int		pos[4];
 
 	pos[0] = event->x_root;
 	pos[1] = event->y_root;
 	pos[2] = MENU_MARGIN(side);
+	/* FIXME: Should we read screen from event's window rather than
+	 * using default? */
+	pos[3] = gdk_screen_get_monitor_at_point(
+				gdk_screen_get_default(),
+				event->x_root, event->y_root);
 
 	icon_prepare_menu((Icon *) pi, FALSE);
 
@@ -2048,10 +2280,10 @@ static void panel_show_menu(GdkEventButton *event, PanelIcon *pi, Panel *panel)
 
 /* Note: also called from icon handler */
 static gboolean panel_drag_motion(GtkWidget	*widget,
-                            GdkDragContext	*context,
-                            gint		x,
-                            gint		y,
-                            guint		time,
+			    GdkDragContext	*context,
+			    gint		x,
+			    gint		y,
+			    guint		time,
 			    Panel		*panel)
 {
 	int panel_x, panel_y;
@@ -2065,10 +2297,10 @@ static gboolean panel_drag_motion(GtkWidget	*widget,
 }
 
 static gboolean insert_drag_motion(GtkWidget	*widget,
-                            GdkDragContext	*context,
-                            gint		x,
-                            gint		y,
-                            guint		time,
+			    GdkDragContext	*context,
+			    gint		x,
+			    gint		y,
+			    guint		time,
 			    Panel		*panel)
 {
 	int panel_x, panel_y;
@@ -2081,7 +2313,7 @@ static gboolean insert_drag_motion(GtkWidget	*widget,
 
 /* Note: also called from icon handler */
 static void panel_drag_leave(GtkWidget	*widget,
-                       GdkDragContext	*context,
+		       GdkDragContext	*context,
 		       guint32		time,
 		       Panel		*panel)
 {
@@ -2093,6 +2325,10 @@ static void panel_drag_leave(GtkWidget	*widget,
 	gdk_window_get_pointer(window, &x, &y, NULL);
 	if (x < 0 || y < 0 || x > alloc->width || y > alloc->height)
 	{
+		keep_below(panel->window->window, TRUE);
+
+		/* Shouldn't need this as well as keep_below but some WMs don't
+		 * automatically lower as soon as the hint is set */ 
 		pinboard = pinboard_get_window();
 		window_put_just_above(panel->window->window, pinboard);
 	}
