@@ -41,6 +41,8 @@ void panel_save(Panel *panel);
 
 gboolean panel_add(PanelSide side,
 		   const gchar *path, const gchar *label, gboolean after);
+gboolean panel_remove_item(PanelSide side, const gchar *path,
+			   const gchar *label);
 void panel_mark_used(GdkRegion *used);
 void panel_update_size(void);
 
