@@ -1,8 +1,8 @@
 /*
- * $Id: pixmaps.c,v 1.114 2004/03/24 20:13:16 kerofin Exp $
+ * $Id: pixmaps.c,v 1.120 2005/01/16 16:48:02 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -83,6 +83,7 @@ MaskedPixmap *im_symlink;
 MaskedPixmap *im_unmounted;
 MaskedPixmap *im_mounted;
 MaskedPixmap *im_appdir;
+MaskedPixmap *im_xattr;
 
 MaskedPixmap *im_dirs;
 
@@ -134,7 +135,7 @@ void pixmaps_init(void)
 
 	pixmap_cache = g_fscache_new((GFSLoadFunc) image_from_file, NULL, NULL);
 
-	gtk_timeout_add(10000, purge, NULL);
+	g_timeout_add(10000, purge, NULL);
 
 	factory = gtk_icon_factory_new();
 	for (i = 0; i < G_N_ELEMENTS(stocks); i++)
@@ -315,7 +316,8 @@ void pixmap_background_thumb(const gchar *path, GFunc callback, gpointer data)
 			    info1.st_dev == info2.st_dev &&
 			    info1.st_ino == info2.st_ino)
 		{
-			pixbuf = gdk_pixbuf_new_from_file(path, NULL);
+			pixbuf = rox_pixbuf_new_from_file_at_scale(path,
+					PIXMAP_THUMB_SIZE, PIXMAP_THUMB_SIZE, TRUE, NULL);
 			if (!pixbuf)
 			{
 				g_fscache_insert(pixmap_cache,
@@ -430,7 +432,9 @@ static void save_thumbnail(const char *pathname, GdkPixbuf *full)
 	smtime = g_strdup_printf("%ld", (long) info.st_mtime);
 
 	path = pathdup(pathname);
-	uri = g_strconcat("file://", path, NULL);
+	uri = g_filename_to_uri(path, NULL, NULL);
+	if (!uri)
+	        uri = g_strconcat("file://", path, NULL);
 	md5 = md5_hash(uri);
 	g_free(path);
 		
@@ -484,7 +488,9 @@ static gchar *thumbnail_path(const char *path)
 	GString *to;
 	gchar *ans;
 	
-	uri = g_strconcat("file://", path, NULL);
+	uri = g_filename_to_uri(path, NULL, NULL);
+	if(!uri)
+	       uri = g_strconcat("file://", path, NULL);
 	md5 = md5_hash(uri);
 		
 	to = g_string_new(home_dir);
@@ -536,7 +542,8 @@ static void child_create_thumbnail(const gchar *path)
 {
 	GdkPixbuf *image;
 
-	image = gdk_pixbuf_new_from_file(path, NULL);
+	image = rox_pixbuf_new_from_file_at_scale(path,
+			PIXMAP_THUMB_SIZE, PIXMAP_THUMB_SIZE, TRUE, NULL);
 
 	if (image)
 		save_thumbnail(path, image);
@@ -581,7 +588,9 @@ static GdkPixbuf *get_thumbnail_for(const char *pathname)
 	struct stat info;
 
 	path = pathdup(pathname);
-	uri = g_strconcat("file://", path, NULL);
+	uri = g_filename_to_uri(path, NULL, NULL);
+	if(!uri)
+	        uri = g_strconcat("file://", path, NULL);
 	md5 = md5_hash(uri);
 	g_free(uri);
 	
@@ -932,6 +941,7 @@ static void load_default_pixmaps(void)
 	im_unmounted = mp_from_stock(ROX_STOCK_MOUNT, mount_icon_size);
 	im_mounted = mp_from_stock(ROX_STOCK_MOUNTED, mount_icon_size);
 	im_appdir = load_pixmap("application");
+	im_xattr = load_pixmap("rox-xattr");
 
 	im_dirs = load_pixmap("dirs");
 
