@@ -1,8 +1,8 @@
 /*
- * $Id: remote.c,v 1.52 2003/10/26 16:14:28 tal197 Exp $
+ * $Id: remote.c,v 1.58 2005/01/16 16:48:02 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2003, the ROX-Filer team.
+ * Copyright (C) 2005, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -95,7 +95,9 @@ static xmlNodePtr rpc_FileType(GList *args);
 static xmlNodePtr rpc_Mount(GList *args);
 
 static xmlNodePtr rpc_PanelAdd(GList *args);
+static xmlNodePtr rpc_PanelRemove(GList *args);
 static xmlNodePtr rpc_PinboardAdd(GList *args);
+static xmlNodePtr rpc_PinboardRemove(GList *args);
 static xmlNodePtr rpc_SetBackdrop(GList *args);
 static xmlNodePtr rpc_SetBackdropApp(GList *args);
 
@@ -123,7 +125,7 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 
 	soap_register("Run", rpc_Run, "Filename", NULL);
 	soap_register("OpenDir", rpc_OpenDir, "Filename",
-					      "Style,Details,Sort,Class,ID");
+			      "Style,Details,Sort,Class,ID,Hidden,Filter");
 	soap_register("CloseDir", rpc_CloseDir, "Filename", NULL);
 	soap_register("Examine", rpc_Examine, "Filename", NULL);
 	soap_register("Show", rpc_Show, "Directory,Leafname", NULL);
@@ -141,7 +143,9 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 	soap_register("SetBackdrop", rpc_SetBackdrop, "Filename,Style", NULL);
 	soap_register("SetBackdropApp", rpc_SetBackdropApp, "App", NULL);
 	soap_register("PinboardAdd", rpc_PinboardAdd, "Path,X,Y", "Label");
+	soap_register("PinboardRemove", rpc_PinboardRemove, "Path", "Label");
 	soap_register("PanelAdd", rpc_PanelAdd, "Side,Path", "Label,After");
+	soap_register("PanelRemove", rpc_PanelRemove, "Side,Path", "Label");
 
 	/* Look for a property on the root window giving the IPC window
 	 * of an already-running copy of this version of the filer, running
@@ -573,6 +577,8 @@ static xmlNodePtr rpc_OpenDir(GList *args)
 {
 	char	   *path;
 	char       *style, *details, *sort, *class, *window;
+	int         hidden=FALSE;
+	char       *filter_string=NULL;
 	FilerWindow *fwin = NULL;
 
 	path = string_value(ARG(0));
@@ -600,6 +606,8 @@ static xmlNodePtr rpc_OpenDir(GList *args)
 	style = string_value(ARG(1));
 	details = string_value(ARG(2));
 	sort = string_value(ARG(3));
+	hidden = bool_value(ARG(6));
+	filter_string = string_value(ARG(7));
 
 	if (style)
 	{
@@ -667,6 +675,15 @@ static xmlNodePtr rpc_OpenDir(GList *args)
 
 		g_free(sort);
 	}
+	if (hidden!=-1)
+	{
+		display_set_hidden(fwin, hidden);
+	}
+
+	if (filter_string)
+		display_set_filter(fwin, FILER_SHOW_GLOB, filter_string);
+	else
+		display_set_filter(fwin, FILER_SHOW_ALL, NULL);
 
 	return NULL;
 }
@@ -787,6 +804,26 @@ static xmlNodePtr rpc_PinboardAdd(GList *args)
 	pinboard_pin(path, name, x, y, NULL);
 
 	g_free(path);
+	if(name)
+	        g_free(name);
+
+	return NULL;
+}
+
+/* args = Path, [Label] */
+static xmlNodePtr rpc_PinboardRemove(GList *args)
+{
+	char *path = NULL;
+	gchar *name;
+
+	path = string_value(ARG(0));
+	name = string_value(ARG(1));
+	
+	pinboard_remove(path, name);
+
+	g_free(path);
+	if(name)
+	        g_free(name);
 
 	return NULL;
 }
@@ -853,6 +890,27 @@ static xmlNodePtr rpc_PanelAdd(GList *args)
 	return NULL;
 }
 
+static xmlNodePtr rpc_PanelRemove(GList *args)
+{
+	PanelSide side;
+	char *path, *side_name, *label;
+
+	side_name = string_value(ARG(0));
+	side = panel_name_to_side(side_name);
+	g_free(side_name);
+	g_return_val_if_fail(side != PANEL_NUMBER_OF_SIDES, NULL);
+
+	path = string_value(ARG(1));
+	label = string_value(ARG(2));
+
+	panel_remove_item(side, path, label);
+
+	g_free(path);
+	g_free(label);
+
+	return NULL;
+}
+
 static xmlNodePtr rpc_Copy(GList *args)
 {
 	GList *from;
@@ -1016,7 +1074,7 @@ static void soap_send(GtkWidget *from, GdkAtom prop, GdkWindow *dest)
 	gdk_event_send_client_message((GdkEvent *) &event,
 				      GDK_WINDOW_XWINDOW(dest));
 
-	gtk_timeout_add(10000, too_slow, NULL);
+	g_timeout_add(10000, too_slow, NULL);
 			
 	gtk_main();
 }
